# global dependencies
import os
import inspect
import numpy as np
import traceback
import json

from time import time, sleep
from copy import deepcopy
from collections import deque, OrderedDict
from functools import partial
from threading import Thread

# local dependencies
from naeural_core import constants as ct
from naeural_core.business.mixins_base.threading import _ThreadingAPIMixin
from naeural_core.main.net_mon import NetworkMonitor

from naeural_core.business.base.base_plugin_biz_loop import _BasePluginLoopMixin
from naeural_core.business.base.base_plugin_biz_api import _BasePluginAPIMixin

from naeural_core.business.mixins_libs import (
  _TimeBinsMixin, _PoseAPIMixin,
  _AlertTrackerMixin
)


from naeural_core import Logger
from naeural_core import DecentrAIObject
from naeural_core.local_libraries import _ConfigHandlerMixin
from naeural_core.business.mixins_base import (
  _AlerterMixin, _ShmMixin,
  _EmailerMixin, _DailyIntervalMixin,
  _DataAPIMixin, _WorkingHoursMixin,
  _IntervalAggregationMixin,
  _UploadMixin, _CmdAPIMixin, _ExecutionAuditMixin,
  _GenericUtilsApiMixin, _DiskAPIMixin, _DeAPIMixin,
  _DatasetBuilderMixin, _StateMachineAPIMixin,
)

from naeural_core.utils.mixins.code_executor import _CodeExecutorMixin

from naeural_core.data_structures import GeneralPayload
from naeural_core.utils.config_utils import get_now_value_from_time_dict
from naeural_core.business.test_framework.testing_manager import TestingManager
from naeural_core.business.test_framework.scoring_manager import ScoringManager

from naeural_core.utils.plugins_base.bc_wrapper import BCWrapper

from naeural_core.ipfs import R1FSEngine

_CONFIG = {

  'INSTANCE_COMMAND': {},  # one time command-trigger
  'INSTANCE_COMMAND_LAST': {},  # data witness of the last INSTANCE_COMMAND
  'INSTANCE_COMMAND_LOG': True,  # log the instance command - default TRUE maybe FALSE in future ?

  'COPY_IN_MAIN_THREAD': False,  # flag to switch between copying data in main thread on in plugin instance thread
  'ENCRYPT_PAYLOAD': False,  # flag to independent toggle payload encryption
  "USE_LOCAL_COMMS_ONLY": False, # flag to switch to local comms only for a particular plugin instance

  'RESEND_LAST_STATUS_ON_IDLE': 0,

  'FORCED_PAUSE': False,
  'DISABLED': False,
  
  'CHAINSTORE_PEERS' : [], # list of peers to be used for chainstore and will enable distribution even to non-whitelisted peers
  

  # set this to 1 for real time processing (data will be lost and only latest data be avail)
  # when PROCESS_DELAY is used this should be either bigger than 1 if we want to have previous data
  # but (more likely) set it to 1 just to "see" current data ("real time at processing time")

  # 'REQUIRES_INPUTS_LOCAL_COPY' TODO: implement safe plugins only - default True for user plugins

  'MAX_INPUTS_QUEUE_SIZE': 1,
  'QUEUE_OVERFLOW_SLEEP_TIME': 0.01,

  'LOG_ON_BLOB': False,

  # "WORKING_HOURS"
  #  this should be one of the following options:
  #     - list of lists (as is)
  #     - dict with "weekday" : list of lists
  #     - dict with "label" : dict( containing 'INTERVALS' : list-of-lists, weekday, params...) dictionary {"i1" : {"INTERVAL" : [start, end], "PARAM1": ...}}
  # ==>> must modify only `working_hours_current_interval` and must add `params_current_interval`
  #
  'WORKING_HOURS_TIMEZONE': None,
  'WORKING_HOURS': [],
  'IGNORE_WORKING_HOURS': False,

  'PLUGIN_LOOP_RESOLUTION': 20,   # proposed loop res
  'FORCED_LOOP_SLEEP': None,  # use predefined sleep time

  'CLOSE_PIPELINE_WHEN_DONE': False,

  'CONFIDENCE_THRESHOLD': 0.2,

  'RESTART_ALERTERS_ON_CONFIG': True,
  'RESTART_SHMEM_ON_CONFIG': True,
  'THREAD_SAFE_DRAWING': True,

  'NO_WITNESS': False,

  'AI_ENGINE': [],
  "INFERENCE_AI_ENGINE_PARAMS": {
  },
  "STARTUP_AI_ENGINE_PARAMS": {
  },

  # alerter zone
  "ALERT_RAISE_CONFIRMATION_TIME": 0,
  "ALERT_LOWER_CONFIRMATION_TIME": 1,
  "ALERT_DATA_COUNT": 2,
  "ALERT_RAISE_VALUE": 0.5,
  "ALERT_LOWER_VALUE": 0.4,
  "ALERT_MODE": 'mean',
  "ALERT_MODE_LOWER": 'max',
  "ALERT_REDUCE_VALUE": False,
  "AUTO_FORCE_LOWER": 0,
  # end alerter zone

  # If RUNS_ONLY_ON_SUPERVISOR_NODE is set to true, the plugin will only work on the supervisor nodes.
  "RUNS_ONLY_ON_SUPERVISOR_NODE": False,

  # if ALLOW_EMPTY_INPUTS is set to true the on-idle will trigger continously the process
  # default is False, set to True if we want to process empty inputs
  'ALLOW_EMPTY_INPUTS': False,

  'SIMPLE_WITNESS': False,  # only the simple picture is returned

  'PROCESS_DELAY': 0,  # minimum time between two consecutive process calls
  'MAX_PROCESS_DELAY': None,  # maximum time between two consecutive process calls


  'ORIGINAL_FRAME': True,
  'RESIZE_WITNESS': [],



  # debug & testing
  'SIMULATE_CRASH_EXEC_ITER': None,  # set this to a nr to simulate a exec crash at that process iter
  'SIMULATE_CRASH_EXEC_SINGLE': True,  # just one crash
  'SIMULATE_CRASH_LOOP': False,  # set this to a nr to simulate a loop crash at that process iter

  'DEBUG_PAYLOADS': False,  # TO BE CLARIFIED

  'DEBUG_MODE': False,


  'DEBUG_SAVE_IMG': False,  # saves in _output/[STREAM]__[SIGN]__[INST] the witness & originals
  'DEBUG_SAVE_IMG_ARCHIVE': 30,     # no file tuples to archive & delete


  # debug save upload migrated from heavy ops
  'DEBUG_SAVE_PAYLOAD': False,  # basic saving of payload as csv in GeneralPayload object

  # next two are not currently implemented
  'DEBUG_SAVE_ARCHIVE_M': 60,  # How often the debug upload should be performed (in minutes)
  'DEBUG_SAVE_UPLOAD': False,

  'TESTING': {},

  'DEBUG_SAVE': False,  # this is left here for legacy/history SaveImageHeavyOps


  'SECONDS_EXPORT': 0,        # REVIEW:: musth check
  # end debug and testing

  # Timebins
  'WEEKDAY_NAMES': ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday'],
  'REPORT_DEFAULT_EMPTY_VALUE': None,
  'PER_DAY_OF_WEEK_TIMESLOT': True,   # True if we want to report day and hour vs hour
  'DAILY_INTERVALS': 144,  # split the day in equal intervals
  'WARMUP_ANOMALY_MODELS': 10,  # how many datapoins required for anomaly model inference
  # end Timebins

  # plugin level data availability control
  'ALERT_ON_NO_DATA': False,
  'ALERT_ON_NO_DATA_INTERVAL': 30,
  'ALERT_ON_NO_DATA_REPEAT': 60,
  # end plugin level data availability control


  # couple of keys for custom/rest plugins manifest sending (and any other plugins that send manifests)
  'SEND_MANIFEST_EACH': 300,
  'LOG_MANIFEST_SEND': False,
  
  # Network Processor
  'ACCEPT_SELF' : False,  
  'FULL_DEBUG_PAYLOADS' : False,
  # END Network Processor  



  'ENABLE_AUDIT': False,

  'TAGS': '',
  'ID_TAGS': [],
  'SAVE_INTERVAL': 5 * 60,   # REVIEW: must check
  'MAX_IDLE_TIME': 5,        # REVIEW: must check

  'LINKED_INSTANCES': [],

  "DATASET_BUILDER": None,

  # SERIALIZATION & PERSISTENCE
  'LOAD_PREVIOUS_SERIALIZATION': False,
  'SERIALIZATION_SIGNATURE': None,
  # END SERIALIZATION & PERSISTENCE


  'VALIDATION_RULES': {
    'FORCED_PAUSE': {
      'DESCRIPTION': 'Will STOP the instance execution until is not `True` anymore',
      'TYPE': 'bool',
    },

    'AUTO_FORCE_LOWER': {
      'DESCRIPTION': 'If non zero will force alerter from Raised to Lowered after the given amount of seconds. WARNING: While this is a generic and universal approach, some plugins have own specific internal lowering mechanics',
      'TYPE': 'float'
    },

    'RESEND_LAST_STATUS_ON_IDLE': {
      'DESCRIPTION': 'If above 0 will resend the instance status (last payload) each given number of seconds from last idle',
      'TYPE': 'float',
    },

    'WORKING_HOURS': {
      'DESCRIPTION': 'Define working hours intervals. Default is full time, otherwise list of [[start, end], [start, end]] strings',
      'TYPE': ['list', 'dict']
    },

    'MAX_INPUTS_QUEUE_SIZE': {
      'DESCRIPTION': "Advanced setting: Controls the queue size of the plugin instance. set this to 1 for real time & overwrite (for example when PROCESS_DELAY is used and we need to process only the 'live' data)",
      'TYPE': 'int',
      'MIN_VAL': 1,
      'MAX_VAL': 512,
    },

    'PROCESS_DELAY': {
      'DESCRIPTION': 'Imposes a specific delay in seconds between business plugin execution cycles - i.e. 5 means cycles are executed each 5 seconds',
      'TYPE': 'float',
      'MIN_VAL': 0,
      'MAX_VAL': 24 * 3600 + 1,  # max delay is one day
    },

    'ALERT_ON_NO_DATA': {
      'DESCRIPTION': 'Automatic alert is raised if the business plugin does not receive data for `ALERT_ON_NO_DATA_INTERVAL` seconds',
      'TYPE': 'bool',
    },

    'ALERT_ON_NO_DATA_INTERVAL': {
      'DESCRIPTION': "Maximal time accepted for a biz plugin to 'work' without data - only valid if `ALERT_ON_NO_DATA = true`",
      'TYPE': 'int',
    },

    'AI_ENGINE': {
      'DESCRIPTION': "AI Engine that will process data upstream from biz plugin. String name or list of strings if multiple AI engines are required.",
      'TYPE': ['str', 'list'],
    },

    'STARTUP_AI_ENGINE_PARAMS': {
      'TYPE': 'dict',
      'DESCRIPTION': 'This configuration dict will be send as startup params during Serving Process creation, thus can overwrite various things such as "DEFAULT_DEVICE"',
    },

    'INFERENCE_AI_ENGINE_PARAMS': {
      'TYPE': 'dict',
      'DESCRIPTION': 'This configuration dict will be added for each individual inference run in "SERVING_PARAMS" key beside "DATA" key during eache `predict`',
    },

    'CLOSE_PIPELINE_WHEN_DONE': {
      'DESCRIPTION': "User flag that can be used to call a `cmdapi_archive_pipeline()`",
      'TYPE': 'bool',
    },

    'SEND_MANIFEST_EACH': {
      "DESCRIPTION": "Idle time that will trigger manifest generation for certain plugins that require this",
      "TYPE": "float"
    },

    'WEEKDAY_NAMES': {
      "DESCRIPTION": "Name of week days, used for the formatted report",
      "TYPE": "list"
    },
    # 'REPORT_DEFAULT_EMPTY_VALUE' : {
    #   # TODO(AID): help decide on this
    #   "DESCRIPTION": "Default value for aggregation function, used for the formatted report when timebins contain no elements",
    #   "TYPE": "obj",
    # },
    'PER_DAY_OF_WEEK_TIMESLOT': {
      "DESCRIPTION": "Controlls if the timebins report contains weekdays or not (on false it only reports intervals, on true it reports intervals by days)",
      "TYPE": "bool"
    },
    'DAILY_INTERVALS': {
      "DESCRIPTION": "Number of intervals in which to split a day (24 means intervals of one hour, 144 means intervals of 10 minutes)",
      "TYPE": "int"
    },
    'WARMUP_ANOMALY_MODELS': {
      "DESCRIPTION": "Number of timesteps required for anomaly model warmup. Defaults to 10",
      "TYPE": "int"
    },

    'LINKED_INSTANCES': {
      "DESCRIPTION": "List of tuple/lists [<STREAM_NAME>,<INSTANCE_ID>] that identify all the 'subordinates' of the current instances",
      "TYPE": "list",
    }
  }
}


class BasePluginExecutor(
  DecentrAIObject,
  _ConfigHandlerMixin,
  _AlerterMixin,
  _ShmMixin,
  _EmailerMixin,
  _DataAPIMixin,
  _DeAPIMixin,
  _ThreadingAPIMixin,
  _WorkingHoursMixin,
  _IntervalAggregationMixin,
  _DailyIntervalMixin,
  _UploadMixin,
  _CmdAPIMixin,
  _ExecutionAuditMixin,
  _GenericUtilsApiMixin,
  _DiskAPIMixin,
  _TimeBinsMixin,
  _CodeExecutorMixin,
  _DatasetBuilderMixin,
  _AlertTrackerMixin,
  _PoseAPIMixin,
  _BasePluginLoopMixin,
  _BasePluginAPIMixin,
  _StateMachineAPIMixin,
):
  CONFIG = _CONFIG

  def __init__(self, log: Logger,
               global_shmem,
               plugins_shmem,
               stream_id,
               signature,
               default_config,
               upstream_config,
               environment_variables=None,
               version="0.0.0",
               initiator_id=None,
               initiator_addr=None,
               session_id=None,
               threaded_execution_chain=True,
               payloads_deque=None,
               commands_deque=None,
               ee_ver=None,
               runs_in_docker=False,
               docker_branch='main',
               debug_config_changes=False,
               pipeline_use_local_comms_only=False,
               pipelines_view_function=None,
               **kwargs):
    self.__version__ = version

    self.__global_shmem = global_shmem  # global access to various engines - not allowed in plugin code

    self.__blockchain_manager = global_shmem[ct.BLOCKCHAIN_MANAGER]  # blockchain manager
    
    self.__r1fs : R1FSEngine = global_shmem[ct.R1FS_ENGINE]  # R1FS (Private IPFS)

    self.__bc = BCWrapper(self.__blockchain_manager, owner=self)  # blockchain wrapper (for encryption/decryption

    self.__plugins_shmem = plugins_shmem  # plugins shared memory

    self.__debug_config_changes = debug_config_changes
    
    self.__pipelines_view_function = pipelines_view_function # view function for this node pipelines

    self._painter = None
    self.__ee_ver = ee_ver
    self.__runs_in_docker = runs_in_docker
    self.__docker_branch = docker_branch
    # TODO: change all to PRIVATE

    self.__stream_id = stream_id
    self.__signature = signature
    self._default_config = default_config
    self._upstream_config = upstream_config
    self._environment_variables = environment_variables
    self.__initiator_id = initiator_id
    self.__initiator_addr = initiator_addr
    self.__modified_by_id = self.__initiator_id  # default modified by is the initiator
    self.__modified_by_addr = self.__initiator_addr  # default modified by is the initiator
    self._session_id = session_id
    self._signature_hash = None
    self._threaded_execution_chain = threaded_execution_chain
    self._timers_section = None
    
    self._init_process_finalized = False

    self._instance_config = None
    
    self.__pipeline_use_local_comms_only = pipeline_use_local_comms_only

    log.P("Init {} v{} <{}:{}> SUPER:{}".format(
      self.__class__.__name__, self.__version__,
      self.initiator_id, self.initiator_addr, self.is_supervisor_node,
    ), color=ct.COLORS.BIZ
    )

    # error handling vars
      

    self._last_data_input = time()  # last time the loop encountered data
    self._last_alert_on_no_data = time()
    self.no_data_alerts_count = 0

    self.lost_inputs_count = 0
    self.queue_full_warning = False
    self.queue_full_delays_count = 0
    self._last_logged_queue_full_count = 0
    self._last_logged_queue_full_time = time()
    # end error handling vars

    # standard payload & process delay variables
    self.__first_process_time = None       # time of the first process call
    self.__last_process_time = None        # tune of the last process call
    self.__last_payload_process_time = 0  # this is the time of last payload creation
    self.__last_payload_resend_time = 0
    self.__last_payload_nonresend_time = 0
    self.__last_saved = time()
    self.__start_time = time()  # time counter since the plugin is alive, usefull for statefull plugins
    self.__total_payloads = 0
    # end delay vars

    # execution vars
    self.done_loop = False

    self._was_stopped_last_iter = False  # for `FORCED_PAUSE`

    self._plugin_loop_in_exec = False

    self.__dct_last_payload = None  # cache of the last payload, starts with nothing

    self._pre_process_outputs = None
    self._payload = None
    self._commands = []  # see cmdapi mixin
    self.payloads_deque = payloads_deque
    self.commands_deque = commands_deque

    if self.payloads_deque is None:
      self.payloads_deque = deque(maxlen=1000)
    if self.commands_deque is None:
      self.commands_deque = deque(maxlen=1000)

    self.__default_payload_data = {}
    # end execution vars

    # END Execution/loop vars

    self.__upstream_inputs_deque = None  # the variable holding the queue
    self.__inputs = None  # the current inputs

    self._testing_manager: TestingManager = None
    self._scoring_manager: ScoringManager = None

    self._dataset_builder_info = None

    self.config_changed_since_last_process = None

    self.thread = None
    super(BasePluginExecutor, self).__init__(log=log, prefix_log='[PLEX]', **kwargs)
    self.init_timestamp = self.log.now_str(nice_print=True, short=False)
    return
  
  @property
  def bc(self):
    """
    This property returns the blockchain manager object that is used by the plugin and expose the following functions:
      - address
      - sign
      - verify
      - encrypt_str
      - decrypt_str
      
    """
    return self.__bc
  
  @property
  def r1fs(self) -> R1FSEngine:
    return self.__r1fs

  @property
  def ee_ver(self):
    return self.__ee_ver

  @property
  def runs_in_docker(self):
    return self.__runs_in_docker

  @property
  def docker_branch(self):
    return self.__docker_branch

  @property
  def _stream_id(self):
    return self.__stream_id

  @property
  def _signature(self):
    return self.__signature

  @property
  def global_shmem(self): #TODO: maybe remove this property
    return self.__global_shmem
  
  @property
  def plugins_shmem(self):
    return self.__plugins_shmem
  
  
  ### Edge Node Pipelines
  
  #TODO: this functions should not be accesible by user-level plugins
  @property
  def node_pipelines(self):
    if self.__pipelines_view_function is not None:
      return self.__pipelines_view_function()
    return None
  
  @property
  def local_pipelines(self):
    return self.node_pipelines
  
  ## End Edge Node Pipelines
  

  @property
  def is_supervisor_node(self):
    """
    Returns `True` if the plugin is running on the supervisor node and `False` otherwise.
    Warning: This property is not safe/trusted while the plugin is initializing due to the 
    fact that the supervisor node may not be set yet within the NET_MON plugin.
    """
    return self.global_shmem.get('is_supervisor_node', False)
  
  @property
  def evm_network(self):
    return self.global_shmem.get('__evm_network', None)

  def __repr__(self):
    s = "<stream='{}' sign='{}' inst='{}'>".format(self._stream_id, self._signature, self.cfg_instance_id)
    return s

  def _reset_plugin_instance_data(self, **kwargs):
    # define logic in inherited class
    self.P("  Plugin instance data reset request received", color='m')
    # REVIEW: basic cleanup
    # ...
    return

  def _reset_plugin_instance_alerters(self):
    self.P("  Plugin instance alerters reset request received", color='m')
    self.alerter_hard_reset_all()
    return

  def reset_plugin_instance(self, **kwargs):
    self.P("Resetting {}".format(self), color='m')
    self._reset_plugin_instance_data(**kwargs)
    self._reset_plugin_instance_alerters()
    return

  def _get_methods(self, cls, include_parent=False):
    return self.log.get_class_methods(cls, include_parent=include_parent)

  def reset_default_plugin_vars(self):
    self.__default_payload_data = {}
    return

  def get_default_plugin_vars(self):
    return self.__default_payload_data

  def payload_set_data(self, key, val):
    self.__default_payload_data[key] = val
    return

  def P(self, s, color=None, **kwargs):
    if color is None or (isinstance(color, str) and color[0] not in ['e', 'r']):
      color = ct.COLORS.BIZ
    super().P(s, prefix=True, color=color, **kwargs)
    return

  def __set_loop_stage(self, s, prefix='2.bm.refresh._chkinst.'):
    str_id = '{}:{}:{}-'.format(self._stream_id, self._signature, self._default_config.get('INSTANCE_ID'))
    self.global_shmem['__set_loop_stage_func'](prefix + str_id + s)
    return

  def startup(self):
    super().startup()

    if False:
      # following code has been left here for legacy reasons
      netmon_class_functions = self.log.get_class_methods(self.network_monitor.__class__, include_parent=False)
      for method_name, func in netmon_class_functions:
        if 'network_' in method_name:
          # make aliases to the network_monitor's methods
          setattr(self, method_name, partial(func, self=self.network_monitor))
      # endif netmon disabled approach

    self.__name__ = self.log.name_abbreviation(self._signature)

    self._instance_init()  # will be moved to thread init
    self.maybe_init_ds_builder_saved_stats()
    return

  def _instance_init(self):
    self.__set_loop_stage('_update_instance_config')
    # now update instance config
    self._update_instance_config()
    # end upteda instance condif
    self.__name__ = 'BP:{}>{}>{}'.format(
      self._stream_id,  # leave it like this for clear logging when receiving parallel jobs
      self.log.name_abbreviation(self._signature),  # can be abbreviated
      self.cfg_instance_id,  # maybe better not abbreviated
    )
    if self._threaded_execution_chain:
      self._timers_section = '{}->{}->{}'.format(self._stream_id, self._signature, self.cfg_instance_id)
    else:
      raise ValueError(
        "Non-threaded plugin instance execution is not supported anymore. Supress temporaly this exception with care")

    instance_full_id = (self._stream_id, self._signature, self.cfg_instance_id)
    self._signature_hash = self.log.hash_object(instance_full_id, size=4)

    if (self.cfg_debug_save or self.cfg_debug_save_img) and not os.path.exists(self.save_path):
      os.makedirs(self.save_path, exist_ok=True)

    if bool(self.cfg_testing):
      self.__set_loop_stage('TestingManager')
      self._testing_manager = TestingManager(log=self.log, owner=self)
      self._scoring_manager = ScoringManager(log=self.log, owner=self)

      try:
        self._testing_manager.create_tester(
          name=self.testing_tester_name,
          config=self.testing_tester_config
        )
      except Exception as e:
        self._testing_manager = None
        self.P("Exception when creating tester:\n{}".format(e), color='r')
        self.P("The plugin will switch from testing mode to normal mode.", color='r')

      try:
        self._scoring_manager.maybe_create_scorer(
          name=self.testing_tester_name,
          y_true_src=self.testing_tester_y_true_src
        )
      except Exception as e:
        self._scoring_manager = None
        self.P("Exception when creating scorer:\n{}".format(e), color='r')
    # endif TESTING == True
    self.__upstream_inputs_deque = deque(maxlen=self.cfg_max_inputs_queue_size)
    self.reset_exec_counter_after_config()
    return

  @property
  def instance_hash(self):
    return self._signature_hash

  @property
  def last_payload_time(self):
    return self.__last_payload_process_time

  @property
  def last_payload_time_str(self):
    return self.log.time_to_str(self.__last_payload_process_time)

  @property
  def total_payload_count(self):
    return self.__total_payloads

  @property
  def input_queue_size(self):
    """
    Returns the size of the input queue that is consumed iterativelly
    """
    return len(self.upstream_inputs_deque)

  # TODO: TO BE REMOVED
  @property
  def cfg_collect_payloads_until_seconds_export(self):
    bool_collect = self._instance_config.get('COLLECT_PAYLOADS_UNTIL_SECONDS_EXPORT', True)
    if self.cfg_seconds_export == 0:
      bool_collect = False
    if bool(self.cfg_testing):
      # the testing plugins should register each individual payload
      bool_collect = False
    return bool_collect

  @property
  def cfg_demo_mode(self):
    if self.cfg_debug_mode:
      return True
    return self._instance_config.get('DEMO_MODE', False)

  @property
  def use_local_comms_only(self):
    return self.cfg_use_local_comms_only or self.__pipeline_use_local_comms_only

  @property
  def is_debug_mode(self):
    return self.cfg_debug_mode

  @property
  def is_demo_mode(self):
    return self.cfg_demo_mode

  def get_plugin_loop_resolution(self):
    # plugin must be faster than main loop (or configured as needed)
    return max(
      self.cfg_plugin_loop_resolution,
      3 * self.global_shmem[ct.CALLBACKS.MAIN_LOOP_RESOLUTION_CALLBACK]()
    )

  @property
  def testing_tester_name(self):
    return self.cfg_testing.get('NAME', None)

  @property
  def testing_tester_y_true_src(self):
    return self.cfg_testing.get('Y_TRUE_SRC', None)

  @property
  def testing_tester_config(self):
    return self.cfg_testing.get('TESTER_CONFIG', {})

  @property
  def testing_scorer_config(self):
    return self.cfg_testing.get('SCORER_CONFIG', {})

  @property
  def testing_upload_result(self):
    return self.cfg_testing.get('UPLOAD_RESULT', False)

  @property
  def cfg_cancel_witness(self):
    return self.cfg_no_witness

  def __maybe_reconfigure_process_delay(self):
    max_process_delay = self.cfg_max_process_delay
    curr_process_delay = self.cfg_process_delay
    if max_process_delay is not None:  # for faster checking
      if isinstance(max_process_delay, (int, float)) and curr_process_delay > max_process_delay:
        self.P("WARNING: PROCESS_DELAY {:.1f} > MAX_PROCESS_DELAY {:.1f} - setting to MAX_PROCESS_DELAY".format(
          curr_process_delay, max_process_delay), color='r',
        )
        self.config_data['PROCESS_DELAY'] = max_process_delay
      # endif
    return

  @property
  def is_process_postponed(self):
    self.__maybe_reconfigure_process_delay()
    _elapsed = self.time_from_last_process
    if _elapsed is None or self.current_process_iteration == 0:
      # if elapsed is none (no __last_process_time recorded) OR no actual process has ever occured then allow run
      return False
    # endif

    if _elapsed < self.cfg_process_delay:
      return True
    # endif elapsed < delay thus wait more until elapsed >= delay
    return False
  
  
  @property
  def is_plugin_temporary_stopped(self):
    msg = None
    status = None
    notif_code = None
    forced_pause = self.cfg_forced_pause
    disabled = self.cfg_disabled
    stopped = forced_pause or disabled
    if self._was_stopped_last_iter and not stopped:
      # just received "resume"
      msg = f"WARNING: Plugin will now RESUME. `FORCED_PAUSE`={forced_pause}, `DISABLED`={disabled}"
      status = "RESUMING"
      notif_code = ct.NOTIFICATION_CODES.PLUGIN_RESUME_OK
    elif not self._was_stopped_last_iter and stopped:
      # just received "stop"
      msg = f"WARNING: Plugin will now STOP. `FORCED_PAUSE`={forced_pause}, `DISABLED`={disabled}"
      status = "PAUSING"
      notif_code = ct.NOTIFICATION_CODES.PLUGIN_PAUSE_OK
      if self.cfg_ignore_working_hours:
        msg_working_hours_conflict = "The following ERROR was detected: IGNORE_WORKING_HOURS is set to True and FORCED_PAUSE is set to True. Although FORCED_PAUSE is above WORKING_HOURS this might be a configuration error."
        msg += ". " + msg_working_hours_conflict
        self.P(msg_working_hours_conflict, color='r')
      # endif working hours conflict
    # endif resume or new stop
    if msg is not None:
      self.P(msg, color='r')
      self._create_notification(
        msg=msg,
        displayed=True,
        forced_pause=self.cfg_forced_pause,
        disabled=self.cfg_disabled,
        status=status,
        notif_code=notif_code,
      )
      self.add_payload_by_fields(
        info=msg,
        forced_pause=self.cfg_forced_pause,
        disabled=self.cfg_disabled,
        status=status,
        notif_code=notif_code,
      )
    # end if process just resumed or just stopped
    self._was_stopped_last_iter = stopped
    return stopped

  @property
  def is_plugin_stopped(self):
    return self.cfg_forced_pause or self.is_outside_working_hours

  @property
  def instance_relative_path(self):
    return '__'.join(self.unique_identification)

  @property
  def save_path(self):
    return os.path.join(self.log.get_output_folder(), self.instance_relative_path)

  @property
  def plugin_output_path(self):
    return self.save_path

  @property
  def last_process_time(self):
    return self.__last_process_time

  @property
  def time_from_last_process(self):
    if self.__last_process_time is None:
      return
    return time() - self.__last_process_time

  @property
  def time_alive(self):
    return time() - self.__start_time

  @property
  def start_time(self):
    return self.__start_time

  @property
  def first_process_time(self):
    return self.__first_process_time

  @property
  def _cache_folder(self):
    return ct.CACHE_PLUGINS

  @property
  def unique_identification(self):
    """
    This property returns a tuple with the following elements:
      - stream_id (id of the pipeline)
      - signature (signature of the plugin)
      - instance_id ()
    
    This tuple is used to uniquely identify the plugin instance in the current node however will
    not be unique in the network as other nodes may have the same plugin instance on a identical 
    pipeline
    """
    return self._stream_id, self._signature, self.cfg_instance_id

  @property
  def plugin_id(self):
    str_name = "{}__{}__{}".format(
      self._stream_id, self._signature, self.cfg_instance_id
    )
    return self.sanitize_name(str_name)

  @property
  def str_unique_identification(self):
    return str(self.unique_identification)
  
  
  @property 
  def full_id(self):
    """
    This property returns the full id of the plugin instance in the following format:
      <node_id>__<stream_id>__<signature>__<instance_id>
    
    This literaly allows the plugin to be uniquely identified in the network.
    """
    str_id = f"{self.e2_addr}__{self._stream_id}__{self._signature}__{self.cfg_instance_id}"
    return str_id

  # set and get inputs
  @property
  def inputs(self):
    return self.__inputs

  @inputs.setter
  def inputs(self, inp):
    self.__inputs = inp
  # end set and get inputs

  @property
  def _device_id(self):
    device_id = self.log.config_data.get(ct.CONFIG_STARTUP_v2.K_EE_ID, '')[:ct.EE_ALIAS_MAX_SIZE]
    return device_id

  @property
  def eeid(self):
    return self._device_id

  @property
  def ee_id(self):
    return self.eeid

  @property
  def node_id(self):
    return self.eeid

  @property
  def e2_id(self):
    return self._device_id

  @property
  def e2_addr(self):
    return self.__blockchain_manager.address

  @property
  def ee_addr(self):
    return self.e2_addr

  @property
  def node_addr(self):
    return self.e2_addr

  @property
  def network_monitor(self) -> NetworkMonitor:
    return self.global_shmem['network_monitor']

  @property
  def netmon(self):
    return self.network_monitor

  @property
  def net_mon(self):
    return self.network_monitor

  @property
  def upstream_inputs_deque(self):
    return self.__upstream_inputs_deque

  @property
  def is_queue_overflown(self):
    # queue is overflown only if max size is bigger than one ("real time") and its at limit
    return (self.input_queue_size >= self.cfg_max_inputs_queue_size) and (self.cfg_max_inputs_queue_size > 1)

  @property
  def time_with_no_data(self):
    return time() - self._last_data_input

  @property
  def modified_by_addr(self):
    return self.__modified_by_addr

  @property
  def initiator_addr(self):
    return self.__initiator_addr

  @property
  def initiator_id(self):
    return self.__initiator_id

  @property
  def modified_by_id(self):
    return self.__modified_by_id

  def get_node_running_time(self):
    """
    Returns the time since the node was started in seconds
    """
    return self.global_shmem['get_node_running_time']()

  def get_node_running_time_str(self):
    """
    Returns the time since the node was started pretty stringyfied
    """
    val = self.global_shmem['get_node_running_time']()
    if val < 600:
      return "{:.0f}s".format(val)
    elif val < 3600:
      return "{:.1f}m".format(round(val / 60, 1))
    else:
      return "{:.1f}h".format(round(val / 3600, 1))

  def add_to_inputs_deque(self, data):
    self.upstream_inputs_deque.append(data)
    return

  def copy_simple_data(self, dct_data):
    dct_res = {}
    for k, v in dct_data.items():
      nl1 = not isinstance(v, (list, str, np.ndarray))
      nl2 = isinstance(v, list) and len(v) > 0 and isinstance(v[0], str) and len(v[0]) < 1000
      nl3 = isinstance(v, (str, np.ndarray)) and len(v) < 1000
      if nl1 and nl2 and nl3:
        dct_res[k] = v
    return deepcopy(dct_res)

  def _cache_last_payload(self, dct_payload):
    self.__dct_last_payload = self.copy_simple_data(dct_payload)
    return

  def __process_payload(self, payload=None):
    """
    This private method wraps the payload-to-dict method and any other stuff such
    as saving the last payload if needed as dict.
    """
    if payload is None:
      payload = self._payload
    if not vars(payload).get('IS_NO_DATA_ALERT', False):
      self._maybe_ds_builder_gather_after_process(payload=payload)
      self._maybe_ds_builder_save()
    dct_res = payload.to_dict()
    self._cache_last_payload(dct_payload=dct_res)
    self.__last_payload_process_time = time()
    return dct_res

  def _maybe_alert_on_no_data(self):
    # this method checks if plugin did not receive data if this plugin so requires
    # and will generate a direct payload with this issue
    if self.cfg_alert_on_no_data:
      if self.time_with_no_data > self.cfg_alert_on_no_data_interval:
        if (time() - self._last_alert_on_no_data) > self.cfg_alert_on_no_data_repeat:
          self.no_data_alerts_count += 1
          self.P("Sending NO DATA alert {}.".format(self.no_data_alerts_count), color='r')
          self._last_alert_on_no_data = time()
          # now we deliver the payload !
          payload = self._create_payload(
            status="No data alert {} - Plugin {} did not receive any data for past {:.0f}s".format(
              self.no_data_alerts_count, self, self.time_with_no_data),
            is_no_data_alert=True,
            no_data_alerts_count=self.no_data_alerts_count,
          )
          self.add_payload(payload)
      # endif interval passed so we maybe raise
    # endif needs to check
    return

  def _maybe_resend_last_payload(self):
    if self.get_last_payload_data() is None:
      return
    if self.cfg_resend_last_status_on_idle > 0:
      if (time() - self.__last_payload_process_time) > self.cfg_resend_last_status_on_idle:
        if self.__last_payload_resend_time != self.__last_payload_process_time:
          # if last process differs last resend then it was not a resend but a normal send so just write it over for below calcs
          self.__last_payload_nonresend_time = self.__last_payload_process_time
        status_payload = self._create_payload(
          status='STATUS PAYLOAD RESEND',
          is_resend=True,
          idle_seconds=round(time() - self.__last_payload_nonresend_time, 2),
          initial_send_time=self.time_to_str(self.__last_payload_nonresend_time),
          resend_last_status_on_idle=self.cfg_resend_last_status_on_idle,
          **self.get_last_payload_data()
        )
        self.add_payload(status_payload)
        self.__last_payload_resend_time = self.__last_payload_process_time
    return

  def _maybe_exec_auto_processes(self):
    """
    This private method executes any required "AUTO_..." processing or other similar stuff.
    Should call mixins if required

    Returns
    -------
    None.

    """
    if self.cfg_auto_force_lower > 0:
      elapsed = self.alerter_time_from_last_change()
      was_reset = self.alerter_maybe_force_lower(max_raised_time=self.cfg_auto_force_lower)
      if was_reset:
        self.add_payload(
          self._create_payload(
            auto_forced_lower=True,
            elapsed_while_raised=round(elapsed, 1),
            status='FORCED LOWER executed after {:.1f}s'.format(elapsed)
              ))
    # end if auto force lower is needed
    return

  def get_last_payload_data(self):
    return self.__dct_last_payload

  def get_plugin_used_memory(self, return_tree=False):
    self.start_timer('get_plugin_memory')
    size_o = self.log.get_obj_size(
      obj=self,
      return_tree=return_tree,
      excluded_obj_props=[
        '_painter', 'thread', 'log', 'owner', 'global_shmem',
        # why excluding - to clarify
      ],
      exclude_obj_props_like=[
        'upstream_inputs_deque'
      ]
    )
    self.end_timer('get_plugin_memory')
    return size_o

  def get_plugin_queue_memory(self):
    self.start_timer('get_plugin_queue_memory')
    size_q = self.log.get_obj_size(obj=self.upstream_inputs_deque)
    self.end_timer('get_plugin_queue_memory')
    return size_q

  def mainthread_wait_for_plugin(self):
    # now we introduce a yield loop to allow the thread to consume if possible
    # WARNING: this approach is bad for the main loop and should NOT be used
    wait_start = time()
    min_decrease = int(self.actual_plugin_resolution)
    wait_time = self.cfg_queue_overflow_sleep_time
    max_wait_time = wait_time * 5
    while self.input_queue_size >= (self.cfg_max_inputs_queue_size - min_decrease):
      sleep(wait_time)
      if (time() - wait_start) > max_wait_time:
        break
    # log post yield-loop status
    if not self.queue_full_warning:
      msg = "    Post sleep queue: {}/{}".format(
        self.input_queue_size,
        self.cfg_max_inputs_queue_size,
      )
      self.P(msg, color='r')
      self.queue_full_warning = True
    return

  def add_inputs(self, inp):
    # RUNNING IN MAIN THREAD (MAIN LOOP)
    if self.is_outside_working_hours:
      # if outside working hours do not add data just skip it
      return

    if not hasattr(self, '_BasePluginExecutor__upstream_inputs_deque'):
      return
    if inp is None:
      return
    if not bool(inp):
      return

    # first things first we have to make a instance specific copy of the original data
    # coming from upstream so that we do not modify data such as numpy by mistake across
    # different parallel instances that use same inputs
    if self.cfg_copy_in_main_thread:
      self.log.start_timer('mainthr_dcopy')  # this is run in main thread
      inp = deepcopy(inp)
      self.log.stop_timer('mainthr_dcopy')

    if self.is_queue_overflown:
      # the following code is executed in the main thread thus it should NOT impose any delays
      # WARNING: if this code is executed then it means there is a problem with the safety measures
      #          implemented within the DCT Manager

      # minimal queue protection triggered only for plugins that queue data and
      # excluding those that need to process just the current timestep (MAX_INPUTS_QUEUE_SIZE = 1)
      self.queue_full_delays_count += 1
      efll = time() - self._last_logged_queue_full_time  # elapsed_from_last_log
      if (not self.queue_full_warning or
              (self._last_logged_queue_full_count != self.queue_full_delays_count and efll > 60)):
        # now we log the issue for maintenance!
        self._last_logged_queue_full_count = self.queue_full_delays_count
        self._last_logged_queue_full_time = time()
        # now we measure queue size
        queue_size_gb = self.get_plugin_queue_memory() / (1024**3)
        plugin_size_gb = self.get_plugin_used_memory() / (1024**3)
        # end measure queue size
        try:
          info = "Queue ovrflw {} (q:{}), Lps {:.1f} c/cfg: {:.1f}/{:.0f}, Q/P: {:.3f}/{:.3f} GB, lost: {}".format(
            self.queue_full_delays_count, self.cfg_max_inputs_queue_size,
            self.actual_plugin_resolution, self.get_plugin_loop_resolution(), self.cfg_plugin_loop_resolution,
            queue_size_gb, plugin_size_gb,
            self.lost_inputs_count,
          )
        except:
          info = "Queue ovrflw {} (q:{}), Lps {} c/cfg: {}/{}, Q/P: {}/{} GB, lost: {}".format(
            self.queue_full_delays_count, self.cfg_max_inputs_queue_size,
            self.actual_plugin_resolution, self.get_plugin_loop_resolution(), self.cfg_plugin_loop_resolution,
            queue_size_gb, plugin_size_gb,
            self.lost_inputs_count,
          )

        self.P(info, color='r')
        PRC_QUEUE_EXCEED_TOTAL_MEMORY = 0.01
        if queue_size_gb > PRC_QUEUE_EXCEED_TOTAL_MEMORY * self.log.total_memory:
          self.P("WARNING: queue exceeds {:.0f}% of total memory size".format(
            PRC_QUEUE_EXCEED_TOTAL_MEMORY * 100), color='r')
        msg = 'Queue overload in {}'.format(self)
        self._create_abnormal_notification(
          msg=msg,
          info=info,
          displayed=True,
        )
      # endif queue-is-full warning required

      # a this point initially a `self.mainthread_wait_for_plugin()` was ran but that imposed main loop unwanted delay
      # as a result the "waiting has been moved to the data acquisition area
    # endif queue is overflown

    # if the queue is full we have to report the lost input
    if self.is_queue_overflown:
      self.lost_inputs_count += 1
    # end record lost inputs
    self.add_to_inputs_deque(inp)
    return

  def start_thread(self):
    self.thread = Thread(
      target=self.plugin_loop,
      args=(),
      name=ct.THREADS_PREFIX + 'plg_' + '>'.join(self.unique_identification),
      daemon=True,
    )
    self.thread.start()
    return

  def stop_thread(self):
    TIMEOUT = 15
    self.P(ct.BM_PLUGIN_END_PREFIX +
           "[main-thr] Received STOP {} thread command...".format(self.instance_hash), color='b')
    self.done_loop = True
    self.thread.join(timeout=TIMEOUT)
    is_alive = self.thread.is_alive()
    self.plugins_shared_mem.clean_instance()
    if not is_alive:
      self.P(ct.BM_PLUGIN_END_PREFIX + "[main-thr] thread {} stopped/joined ok.".format(self.instance_hash), color='b')
    else:
      self.P(ct.BM_PLUGIN_END_PREFIX +
             "[main-thr] Cannot gracefully stop thread {} after {}s".format(self.instance_hash, TIMEOUT), color='r')
    return

  def get_stream_id(self):
    return self._stream_id

  def get_signature(self):
    return self._signature

  def get_instance_id(self):
    return self.cfg_instance_id

  def get_instance_config(self):
    return self._instance_config

  def get_upstream_config(self):
    return self._upstream_config

  def get_payload_after_exec(self):
    """Gets the payload and resets the internal _payload protected variable

    Returns
    -------
    payload : GenericPayload
        returns the current payload
    """
    p = self._payload
    self._payload = None
    return p


  def start_timer(self, tmr_id):
    if self.log.is_main_thread:
      tmr_id = tmr_id + '_{}'.format(self._signature_hash)

    self.log.start_timer(tmr_id, section=self._timers_section)
    return

  def end_timer(self, tmr_id, skip_first_timing=False, periodic=False):
    if self.log.is_main_thread:
      tmr_id = tmr_id + '_{}'.format(self._signature_hash)

    return self.log.end_timer(tmr_id, skip_first_timing=skip_first_timing, section=self._timers_section, periodic=periodic)

  def stop_timer(self, tmr_id, skip_first_timing=False, periodic=False):
    self.end_timer(tmr_id=tmr_id, skip_first_timing=skip_first_timing, periodic=periodic)
    return
  

  def __on_config(self):
    """
    Called when the instance has just been reconfigured

    Parameters
    ----------
    None

    Returns
    -------
    None.

    """
    self._on_config()
    if not self._init_process_finalized and not self.cfg_disabled :
      self.P("Plugin was disabled initially. Running initialization.")
      self._on_init()
      self._init_process_finalized = True
    #endif
    return
    

  def maybe_update_instance_config(self, upstream_config, session_id=None, modified_by_addr=None, modified_by_id=None):
    """
    This method is called by the plugin manager when the instance config has changed.
    IMPORTANT: it runs on the same thread as the BusinessManager so it should not block!

    For the particular case when only INSTANCE_COMMAND is modified then the plugin should not reset its state
    """
    # first cleanup all the non-actionable keys
    for k in self.ct.NON_ACTIONABLE_INSTANCE_CONFIG_KEYS:
      upstream_config.pop(k, None)
      self._upstream_config.pop(k, None)
    
    if upstream_config == self._upstream_config:
      return
    
    # if is just a INSTANCE_COMMAND then bypass the full config update including
    # `is_instance_command_only` check from the `_update_instance_config`
    # first set _upstream_config

    if session_id is not None and session_id != self._session_id:
      self.P("  Changing session_id from '{}' to '{}'".format(self._session_id, session_id), color='y')
      self._session_id = session_id
    if modified_by_id is not None and modified_by_id != self.__modified_by_id:
      self.P("  Changing modified-by-id from '{}' to '{}'".format(self.__modified_by_id, modified_by_id), color='y')
      self.__modified_by_id = modified_by_id
      self.__modified_by_addr = modified_by_addr
    # endif session_id

    # check if is just a command
    if upstream_config.get('INSTANCE_COMMAND') not in [None, '', [], {}]:
      # if the command is not empty then we just set the command and return
      # TODO(FIXME): when a plugin instance `process` method takes some time,
      # this text will be spammed in the logs. also, there is the risk of 
      # a new command overwriting the previous one before it is processed. 
      self.P("Command '{}' received for plugin {}".format(upstream_config['INSTANCE_COMMAND'], self))
      self.config_data['INSTANCE_COMMAND'] = upstream_config['INSTANCE_COMMAND']
      # reset the command for the upstream config
      self._upstream_config['INSTANCE_COMMAND'] = {}  
      upstream_config['INSTANCE_COMMAND'] = {}
      return
    # endif is just a command

    self._upstream_config = upstream_config

    self.P("Config {} on request {}:{}".format(self, modified_by_id, modified_by_addr), color='b')
    # DEBUG
    # self.log.P("Current:\n{}".format(json.dumps(self._upstream_config, indent=4)), color='d')
    # self.log.P("New:\n{}".format(json.dumps(upstream_config, indent=4)), color='d')
    # END DEBUG
    self.config_changed_since_last_process = True

    
    # now while changing config we must stop loop exec
    self.__set_loop_stage(s='maybe_update_instance_config.wait', prefix='2.bm.refresh.{}'.format(self.cfg_instance_id))
    wait_start = self.time()
    CONFIG_WAIT_FOR_EXEC = 10
    while self._plugin_loop_in_exec:
      sleep(0.001)
      if (self.time() - wait_start) >= CONFIG_WAIT_FOR_EXEC:
        msg = "Forced update config will be executed. Plugin is taking too long in `_plugin_loop_in_exec`"
        self.P(msg, color='error')
        self._create_error_notification(
          msg=msg,
          displayed=True,
        )
        break
      # endif need to break the loop
    # end wait loop
    self.__set_loop_stage(s='maybe_update_instance_config.done-wait',
                          prefix='2.bm.refresh.{}'.format(self.cfg_instance_id))

    # pause the plugin loop while updating the config
    self.loop_paused = True  # can also use self.pause_loop() but is safer like this as here we have getter and setter (harder to overwrite)
    last_config = deepcopy(self.config_data)
    try:
      self.__set_loop_stage(s='maybe_update_instance_config._update_instance_config',
                            prefix='2.bm.refresh.{}'.format(self.cfg_instance_id))
      self._update_instance_config()
      # call on-config handler
      self.__set_loop_stage(s='maybe_update_instance_config._on_config',
                            prefix='2.bm.refresh.{}'.format(self.cfg_instance_id))
      self.__on_config()
      # resume loop
      self.__set_loop_stage(s='maybe_update_instance_config.reset_exec_counter_after_config',
                            prefix='2.bm.refresh.{}'.format(self.cfg_instance_id))
      self.reset_exec_counter_after_config()
      self.loop_paused = False  # in the rare case that someone or something overwrote resume_loop()

    except Exception as exc:
      # rollback
      self._instance_config = last_config
      self.config_data = last_config
      msg = "Exception occured while updating the instance config: '{}' Rollback to the last good config.".format(exc)
      info = traceback.format_exc()
      self.P(msg + '\n' + info, color='r')

      # self.__save_full_config()
      # Instead of saving the full config, we only need to rollback the keys that were changed
      # which are in the upstream_config
      # We also need to exclude all the keys that are not in config data, as they are not used by anyone
      self.__save_config(keys=[k for k in upstream_config.keys() if k in self.config_data])

      self._create_error_notification(
        msg=msg,
        info=info,
        displayed=True,
      )
      # resume loop even on exception
      self.loop_paused = False
    return

  def _create_notification(self, msg, info=None, notif=ct.PAYLOAD_CT.STATUS_TYPE.STATUS_NORMAL, use_local_comms_only=False, **kwargs):
    return super()._create_notification(
      notif=notif, msg=msg, info=info,
      stream_name=self._stream_id,
      # next special info for backend
      signature=self._signature,
      instance_id=self.cfg_instance_id,
      initiator_id=self.__initiator_id,
      session_id=self._session_id,
      ct=ct,
      use_local_comms_only=use_local_comms_only or self.use_local_comms_only,
      **kwargs
    )

  def _create_error_notification(self, msg, info=None, **kwargs):
    return super()._create_notification(
      notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
      msg=msg, info=info,
      autocomplete_info=True,
      stream_name=self._stream_id,

      # next special info for backend
      signature=self._signature,
      instance_id=self.cfg_instance_id,
      initiator_id=self.__initiator_id,
      session_id=self._session_id,
      ct=ct,
      **kwargs
    )

  def _create_abnormal_notification(self, msg, info=None, **kwargs):
    return super()._create_notification(
      notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING,
      msg=msg, info=info,
      autocomplete_info=True,
      stream_name=self._stream_id,
      # next special info for backend
      signature=self._signature,
      instance_id=self.cfg_instance_id,
      initiator_id=self.__initiator_id,
      session_id=self._session_id,
      ct=ct,
      **kwargs
    )

  def __save_config(self, keys):
    # writes the config for a set of particular keys on the local cache!
    save_config_fn = self.global_shmem[ct.CALLBACKS.INSTANCE_CONFIG_SAVER_CALLBACK]
    try:
      config = {k: self.config_data[k] for k in keys}
      save_config_fn(self._stream_id, self._signature, self.cfg_instance_id, config_data=config)
      # the save will trigger a update config when the BusinessManager will get the new config from
      # the ConfigManager and will pass it as upstream_config to the plugin
      # so we already update the upstream_config to avoid this
      for k in keys:
        self._upstream_config[k] = self.config_data[k]
    except Exception as exc:
      self.P("Error '{}' while saving keys {}".format(exc, keys))
      raise exc
    return

  def __save_full_config(self):
    # writes the config on the local cache!
    save_config_fn = self.global_shmem[ct.CALLBACKS.INSTANCE_CONFIG_SAVER_CALLBACK]
    try:
      save_config_fn(self._stream_id, self._signature, self.cfg_instance_id, config_data=self.config_data)
      # the save will trigger a update config when the BusinessManager will get the new config from
      # the ConfigManager and will pass it as upstream_config to the plugin
      # so we already update the upstream_config to avoid this
      self._upstream_config = deepcopy(self.config_data)
    except Exception as exc:
      self.P("Error '{}' while saving full config".format(exc))
      raise exc
    return

  def _reset_last_process_time(self):
    self.__last_process_time = time()
    return

  def reset_first_process(self):
    self.__last_process_time = None
    return

  def need_refresh(self):
    idle_time = time() - self.__last_payload_process_time
    return idle_time > self.cfg_max_idle_time

  # TODO: TO BE REMOVED

  def maybe_save_data(self, data, single_file=False):
    """
    DEPRECATED - please use the persistence API
    """
    saved_fn = None
    if (time() - self.__last_saved) > self.cfg_save_interval:
      self.P("[DEPRECATED - please use the persistence API] {} saving its data (SAVE_INTERVAL={}s).".format(
        self.__class__.__name__, self.cfg_save_interval), color='y')
      saved_fn = "{}_{}".format(self._signature, self.log.now_str()) if not single_file else self._signature
      self.log.save_pickle_to_output(
        data=data,
        fn=saved_fn
      )
      self.__last_saved = time()
    return saved_fn

  # Section 0 - methods that provides main functionality of a plugin
  if True:

    def archive_config_keys(self, keys: list, defaults=None):
      """
      Method that allows resetting of a list of keys and saving the current value as `_LAST` keys

      Parameters
      ----------
      keys : list
        List of keys to be archived.

      defaults: list
        List of default values for all keys. Default is None

      Returns
      -------
      None.

      """
      if defaults is None:
        defaults = [None] * len(keys)
      assert len(defaults) == len(keys), "Default values must be provided for all keys"
      all_keys = keys.copy()
      archived_keys = []
      for i, key in enumerate(keys):
        archive_key = key + '_LAST'
        self.config_data[archive_key] = self.config_data[key]
        self.config_data[key] = defaults[i]
        all_keys.append(archive_key)
      # endfor
      self.save_config_keys(all_keys)
      return

    def save_config_keys(self, keys: list):
      """
      Method that allows saving the local config in local cache in order to update a
      specific set of given keys that might have been modified during runtime

      Parameters
      ----------
      keys : list
        List of keys to be saved.

      Returns
      -------
      None.

      """
      EXCEPTIONS = ['INSTANCE_ID']
      result = False
      all_exist = all([k in self.config_data for k in keys])
      all_accepted = all([k not in EXCEPTIONS for k in keys])
      if all_exist and all_accepted:
        self.__save_config(keys)
        result = True
      return result

    def add_payload(self, payload):
      """
      Adds a payload in the plugin instance output queue. If used inside plugins
      plese do NOT return the payload from _process as the payload will be duplicated

      Parameters
      ----------
      payload : GeneralPayload or dict
        the payload

      Returns
      -------
      None.

      """
      if payload is not None:
        # if the payload is added from normal process output then is already __process_payload-ed from high_level_execution_chain
        if isinstance(payload, GeneralPayload):
          # TODO: some things may be missing (cvpluginexecutor._add_plugin_identifiers_to_payload)
          # self._add_plugin_identifiers_to_payload(payload=payload)
          payload = self.__process_payload(payload)
        self.payloads_deque.append(payload)
        self.__total_payloads += 1
      return payload

    def add_payload_by_fields(self, **kwargs):
      """
      Adds a payload in the plugin instance output queue based on fields rather than
      on a already created payload object.
      If used inside plugins plese do NOT return the payload from _process as the payload
      will be duplicated

      Parameters
      ----------
      **kwargs : dict

      Returns
      -------
      None.

      """
      payload = self._create_payload(**kwargs)
      self.add_payload(payload)
      return
    
    def send_encrypted_payload(self, node_addr, **kwargs):
      """
      Sends an encrypted payload to a specific node address

      Parameters
      ----------
      node_addr : str
        The node address to send the payload to
        
      **kwargs : dict
        The payload fields to send

      Returns
      -------
      None.

      """
      payload = {
        **kwargs,
        self.const.PAYLOAD_DATA.EE_IS_ENCRYPTED : True,    
        self.const.PAYLOAD_DATA.EE_DESTINATION : node_addr,  
      }
      self.add_payload_by_fields(**payload)
      return
    
    
    def add_encrypted_payload_by_fields(self, node_addr, **kwargs):
      """
      Sends an encrypted payload to a specific node address. Alias for `send_encrypted_payload`

      Parameters
      ----------
      node_addr : str
        The node address to send the payload to
        
      **kwargs : dict
        The payload fields to send

      Returns
      -------
      None.
      """      
      self.send_encrypted_payload(node_addr=node_addr, **kwargs)
      return
    

    def create_and_send_payload(self, **kwargs):
      """
      Creates a payload and sends it to the output queue.
      If used inside plugins plese do NOT return the payload from _process as the payload
      will be duplicated

      Parameters
      ----------
      **kwargs : dict

      Returns
      -------
      None.

      """
      payload = self._create_payload(**kwargs)
      self.add_payload(payload)
      return

    def get_next_avail_input(self):
      result = None
      if self.input_queue_size > 0:
        data = self.upstream_inputs_deque.popleft()
        if not self.cfg_copy_in_main_thread:
          self.start_timer('thread_dcopy')
          result = deepcopy(data)
          self.stop_timer('thread_dcopy')
        else:
          result = data
      return result

    def pre_process_wrapper(self):
      self.start_timer(ct.TIMER_PRE_PROCESS)
      self._pre_process_outputs = self._pre_process()
      self.stop_timer(ct.TIMER_PRE_PROCESS)
      return

    def process_wrapper(self):
      self.start_timer('process_wrapper')

      if self.__first_process_time is None:
        self.__first_process_time = time()

      self._recalc_plugin_resolution()

      self._reset_last_process_time()  # reset time as we call it next
      self.start_timer('process')
      self._payload = self._process()
      self.stop_timer('process')
      self.stop_timer('process_wrapper')
      return

    def post_process_wrapper(self):
      self.start_timer(ct.TIMER_POST_PROCESS)
      self._post_process()

      if self._payload is not None:
        pass

      self.stop_timer(ct.TIMER_POST_PROCESS)
      return

    # TESTING / BENCHMARKING

    def _maybe_register_payload_for_tester(self, payload=None):
      # this is most likely redundant - should be evaluated for deletion
      if self._testing_manager is None:
        return

      payload = payload or self._payload

      self._prepare_payload_for_testing_registration(payload)
      self._testing_manager.tester.register_payload(payload)
      self._payload = None  # during testing there is no need to communicate the payloads
      return

    def _maybe_generate_testing_results_payload(self, payload):
      if self._testing_manager is None:
        return

      payload = self._prepare_payload_for_generating_testing_results(payload)

      if payload is not None:
        payload['STREAM'] = self.dataapi_stream_name()
        payload['SIGNATURE'] = self.get_signature()
        payload['INSTANCE_ID'] = self.cfg_instance_id
        payload['VERSION'] = self.__version__
        payload['BOX_SESSION_ID'] = self.log.file_prefix
        payload['TESTER_NAME'] = self.testing_tester_name

        if self._scoring_manager is not None:
          dct_score = self._scoring_manager.score(
            y_true_src=self.testing_tester_y_true_src,
            payload=payload,
            config=self.testing_scorer_config
          )

          payload['Y_TRUE'] = self._scoring_manager.get_y_true(
            y_true_src=self.testing_tester_y_true_src,
            name=payload['TESTER_NAME']
          )
          payload['TOTAL_FRAMES_PROCESSED'] = self.current_process_iteration
          payload['SCORE'] = dct_score
        # endif

        file_name = '{}__{}__{}__{}.json'.format(
          payload['BOX_SESSION_ID'],
          payload['STREAM'],
          payload['SIGNATURE'],
          payload['INSTANCE_ID']
        )
        file_path = self.log.save_output_json(
          payload,
          file_name,
          subfolder_path='testing'
        )

        if self.testing_upload_result:
          target_path = '/TESTING/{}'.format(file_name)

          url, _ = self.upload_file(
            file_path=file_path,
            target_path=target_path,
            force_upload=True,
            verbose=1
          )

          payload['URL'] = url

          os.remove(file_path)
        # endif
      # endif
      return

  # endif

  # Section A - methods that can be overwritten in subclass.
  # When reimplementing, pay double attention if you want to overwrite the entire functionality;
  # otherwise, first line should be parent_result = super().method_name(...)
  if True:

    def _create_payload(self, **kwargs):
      if 'plugin_category' not in kwargs:
        kwargs['plugin_category'] = 'general'

      if self.ct.PAYLOAD_DATA.EE_IS_ENCRYPTED.lower() not in kwargs and self.ct.PAYLOAD_DATA.EE_IS_ENCRYPTED.upper() not in kwargs:
        kwargs[self.ct.PAYLOAD_DATA.EE_IS_ENCRYPTED] = self.cfg_encrypt_payload

      payload = GeneralPayload(
        owner=self,
        _p_dataset_builder_used=False,
        **kwargs
      )
      return payload

    def _create_output(self, **kwargs):
      # _create_payload alias
      return self._create_payload(**kwargs)

    def _create_result(self, **kwargs):
      # _create_payload alias
      return self._create_payload(**kwargs)

    def _create_response(self, **kwargs):
      # _create_payload alias
      return self._create_payload(**kwargs)

    def _prepare_payload_for_testing_registration(self, payload):
      return payload

    def _prepare_payload_for_generating_testing_results(self, payload):
      self._payload = None
      return payload

    def _check_delta_config(self, delta_config):
      # see what keys differ from current config - particularly useful for INSTANCE_COMMAND
      updates = []
      for k, v in delta_config.items():
        if (
          (self._instance_config is None) or
          (k not in self._instance_config) or
          (v != self._instance_config[k])
        ):
          updates.append(k)
      return updates

    def _update_instance_config(self):
      if self._instance_config is not None:
        reconfig = True
        self.P("* * * * Reconfiguring plugin {} * *".format(self), color='b')
      else:
        self.P("* * * * * Initial config of plugin `{}` * * *".format(self.__class__.__name__), color='m')
        reconfig = False

      if self._environment_variables is not None and len(self._environment_variables) > 0:
        if self.__debug_config_changes:
          self.P("Updating instance config with node environment config....", color='m')
        # initially this call did not have `default_config = self._instance_config` that led
        # to wrongly using at each update the _default_config
        self.__set_loop_stage('_update_instance_config._environment_variables')
        self._instance_config = self._merge_prepare_config(
          default_config=self._instance_config,
          delta_config=self._environment_variables,
          debug=self.__debug_config_changes,
        )
      # now normal delta config from upstream
      self.__set_loop_stage('_update_instance_config._default_config')

      updates = self._check_delta_config(self._upstream_config)  # check if there are any updates
      is_instance_command_only = len(updates) == 1 and 'INSTANCE_COMMAND' in updates

      self._instance_config = self._merge_prepare_config(
        default_config=self._instance_config,
        debug=self.__debug_config_changes,
      )

      self.__set_loop_stage('_update_instance_config.setup_config_and_validate')

      # if we do not run the next line then the config will not be updated for each of the cfg_* handlers
      # as they will point to the old config_data dict
      self.setup_config_and_validate(self._instance_config)  # here _instance_config will be copyed to config_data

      self.__set_loop_stage('_update_instance_config._print_warnings')
      self._print_warnings()

      # now for the second part of the config: reset/set stuff
      if is_instance_command_only:
        self.P(">>>> INSTANCE_COMMAND received, skipping all resets and initializations... <<<<")
      else:
        # now maybe restart alerters
        if (self.cfg_restart_alerters_on_config and reconfig) or not reconfig:
          # either create if reset is needed or first config (not reconfig)
          # TODO: make sure you set RESTART_ALERTERS_ON_CONFIG to False for plugins
          #       that should NOT restart alerters
          self.__set_loop_stage('_update_instance_config._create_alert_state_machine')
          self._create_alert_state_machine()

        # Create time bins mixin default # TODO: refactor similar with alerters
        self.__set_loop_stage('_update_instance_config.timebins_create_bin')
        self.timebins_create_bin()

        # now maybe re-init plugin shmem if uptate or new initialization
        if (self.cfg_restart_shmem_on_config and reconfig) or not reconfig:
          # `global_shmem` here is WRONG: we should NOT allow  plugins to access global shmem!
          #  so instead we use `plugins_shmem` initiated by the BizMgr
          self.__set_loop_stage('_update_instance_config.init_plugins_shared_memory')
          self.init_plugins_shared_memory(self.plugins_shmem)

      self.last_config_timestamp = self.log.now_str(nice_print=True, short=False)
      if self.is_plugin_stopped:
        self._create_notification(
          msg='Plugin configured while being stopped',
          notif_code=ct.NOTIFICATION_CODES.PLUGIN_CONFIG_IN_PAUSE_OK,
        )
      self.__set_loop_stage('_update_instance_config.EXIT_update_instance_config')
      return

    def _prepare_debug_save_payload(self):
      # this method is "connected" with "debug_save" heavy ops plugin
      return

    def _get_alerter_reduce_threshold(self):
      return self._get_confidence_threshold()

    def _print_warnings(self):
      str_warnings = ""

      # TODO: Maybe remove
      if self.cfg_debug_payloads:
        str_warnings += "- sends payloads with increased frequency for DEBUG purposes."
      ####
      ####

      if not hasattr(self, '_draw_witness_image'):
        str_warnings += "\n- does not have a `_draw_witness_image` method to call from `get_witness_image`"

      _pre_process_ow = inspect.getsource(self._pre_process) != inspect.getsource(BasePluginExecutor._pre_process)
      _process_ow = inspect.getsource(self._process) != inspect.getsource(BasePluginExecutor._process)
      _post_process_ow = inspect.getsource(self._post_process) != inspect.getsource(BasePluginExecutor._post_process)

      if not _process_ow:
        str_warnings += "\n- does not provide high level '_process' functionality. Make sure that _on_data and _on_idle to be overwritten!"

      if not _pre_process_ow:
        str_warnings += "\n- does not provide high level '_pre_process' functionality."

      if not _post_process_ow:
        str_warnings += "\n- does not provide high level '_post_process' functionality"

      if str_warnings != "":
        str_warnings = "{} WARNINGS:".format(self.__class__.__name__) + str_warnings
        self.P(str_warnings, color='y')

      return

    def _witness_prepare(self, img=None, **kwargs):
      img_witness = None
      if img is not None:
        _H, _W = img.shape[:2]
        if self.cfg_original_frame:
          img_witness = img[:, :, ::-1]
          img_witness = np.ascontiguousarray(img_witness, dtype=np.uint8)
        else:
          img_witness = np.zeros((_H, _W, 3), dtype=np.uint8)
      # endif
      return img_witness

    def _witness_pre_process(self, img_witness, **kwargs):
      return img_witness

    def _witness_post_process(self, img_witness, **kwargs):
      return img_witness

    def _add_alert_info_to_payload(self, payload):
      vars(payload)[ct.IS_ALERT] = self.alerter_is_alert()
      vars(payload)[ct.STATUS_CHANGED] = self.alerter_status_changed()
      return

    def _get_confidence_threshold(self):
      """
      Function to generate confidence threshold

      Returns
      -------
      conf_thr : float (0-1)
      """
      conf_thr = self.cfg_confidence_threshold
      if conf_thr is not None:
        if isinstance(conf_thr, dict):
          # assume conf thr is hourly based and keys are "HH:MM-HH:MM" format
          dct_hours = conf_thr
          conf_thr = get_now_value_from_time_dict(dct_hours)
      return conf_thr

    def maybe_start_thread_safe_drawing(self, name='_draw_witness_image'):
      if self.cfg_thread_safe_drawing:
        self._thread_safe_drawing_lock_name = name
        self.log.lock_resource(self._thread_safe_drawing_lock_name)
      return

    def maybe_stop_thread_safe_drawing(self):
      if self.cfg_thread_safe_drawing:
        self.log.unlock_resource(self._thread_safe_drawing_lock_name)
      return

    def get_witness_image(self, img=None,
                          prepare_witness_kwargs=None,
                          pre_process_witness_kwargs=None,
                          draw_witness_image_kwargs=None,
                          post_process_witness_kwargs=None):
      """
      This is the wrapper function that should be called from any plug-in.
      It contains the channel reversing and the cv2 required numpy magic and it
      will call the `_draw_witness_image` plug-in specific method

      definition of: _draw_witness_image(img_witness, **kwargs)

      Parameters
      ----------
      img: np.ndarray
        The starting image. Can be None

      prepare_witness_kwargs : dict
        anything we need in _witness_prepare (passed as **prepare_witness_kwargs)

      pre_process_witness_kwargs : dict
        anything we need in _witness_pre_process (passed as **pre_process_witness_kwargs)

      draw_witness_image_kwargs : dict
        anything we need in _draw_witness_image (passed as **draw_witness_image_kwargs)

      post_process_witness_kwargs : dict
        anything we need in _witness_post_process (passed as **post_process_witness_kwargs)

      Returns
      -------
      img_witness : ndarray
        The returned image will be in RGB format.

      """
      if self.cfg_cancel_witness:
        return

      prepare_witness_kwargs = prepare_witness_kwargs or {}
      pre_process_witness_kwargs = pre_process_witness_kwargs or {}
      draw_witness_image_kwargs = draw_witness_image_kwargs or {}
      post_process_witness_kwargs = post_process_witness_kwargs or {}

      if img is not None:
        if not isinstance(img, np.ndarray):
          self.P("`get_witness_image` called having a non numpy img {}. Returning no witness".format(
            repr(self), type(img)), color='error')
          return

      uses_BGR = False

      self.start_timer(ct.TIMER_GET_WITNESS_IMAGE)
      img_witness = self._witness_prepare(img, **prepare_witness_kwargs)
      if img_witness is not None:
        img_witness = self._witness_pre_process(img_witness, **pre_process_witness_kwargs)
        uses_BGR = True

      if hasattr(self, '_draw_witness_image'):
        if not self.cfg_simple_witness:
          self.maybe_start_thread_safe_drawing()
          img_witness = self._draw_witness_image(
            img=img_witness,
            **draw_witness_image_kwargs
          )
          self.maybe_stop_thread_safe_drawing()

      if img_witness is not None:
        img_witness = self._witness_post_process(img_witness, **post_process_witness_kwargs)
        # maybe we need to resize the whole thing so that is respects the model input size
        if len(self.cfg_resize_witness) == 2:
          h, w = self.cfg_resize_witness
          img_witness = self.log.center_image2(
            np_src=img_witness,
            target_h=h,
            target_w=w
          )

        # because we flipped channels in `_witness_prepare` to BGR in order to draw using cv2; now we come back to RGB..
        if uses_BGR:
          img_witness = img_witness[:, :, ::-1]
      # endif

      self.stop_timer(ct.TIMER_GET_WITNESS_IMAGE)
      return img_witness
  # endif



import abc

from threading import Thread
from naeural_core import constants as ct

from naeural_core.serving.base.base_serving_process import ModelServingProcess as BaseServingProcess

__VER__ = '0.1.0.0'

_CONFIG = {
  **BaseServingProcess.CONFIG,
  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },
}

class ContinousServingProcess(BaseServingProcess):

  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    if False and kwargs['inprocess']:
      raise ValueError("ContinousServingProcess cannot run inprocess")
    self._continous_process_done = False
    self._continous_thread = None
    super(ContinousServingProcess, self).__init__(**kwargs)
    return

  def _startup(self):
    self._continous_thread = Thread(
      target=self.__continous_process,
      args=(),
      name=ct.THREADS_PREFIX + self.__class__.__name__,
      daemon=True,      
    )
    self._continous_thread.start()
    return

  def _pre_process(self, inputs):
    return inputs

  def _predict(self, inputs):
    dummy_results = self._on_status(inputs)
    return dummy_results

  def _post_process(self, preds):
    result = preds.tolist()
    return result
  
  def __continous_process(self):
    while not self._continous_process_done:
      # if self._process is single pass (such as a training process) it must set _continous_process_done to True
      self._process()
    return

  # below must be defined in child class
  def _on_status(self, inputs):
    return inputs

  @abc.abstractmethod
  def _process(self):
    raise NotImplementedError()

  


import numpy as np
import psutil
import threading
import sys
import os

from collections import deque
from time import sleep, time, localtime, strftime

from naeural_core import constants as ct


from naeural_core import DecentrAIObject

__VER__ = '0.5.0'

class SysMonCt:
  PID = 'pid'
  NAME = 'name'
  UTIME = 'usr_time'
  STIME = 'sys_time'
  P_PRC = 'cpu'
  P_TIME = 'total'
  P_CUSR = 'children_user'
  P_CSYS = 'children_system'
  P_MB = 'memory_mb'
  P_MEM = 'memory_info'
  P_CWD = 'cwd'
  P_USR = 'username'
  
  LAST_SEEN = 'last_seen'

  K_THR = 'threads'
  K_CHL = 'CHILDREN'
  K_ID  = 'id'

  # attributes accpted by process inspection  
  P_ATTRS = [
    NAME,
    P_CWD,
    P_USR,
    'status',
    'cmdline',
  ]
  
  NON_TIMESERIES = [
    LAST_SEEN,
  ]
   


class SystemMonitor():
  
  def __init__(self, 
               monitored_prefix, 
               log, 
               max_size=1000, 
               tick=10, 
               DEBUG=False,
               run_processes_each=10,
               ping_interval=300,
               name='S_SysMon',
               **kwargs):
    super(SystemMonitor, self).__init__(**kwargs)    
    self.name = name
    self.monitored_prefix = monitored_prefix
    self.done = True
    self._DEBUG = DEBUG
    self.__version__ = __VER__
    self.log = log
    self.max_size = max_size
    self.main_proc = psutil.Process() 
    self.main_pid = self.main_proc.pid
    self.tick = tick
    self._loop_count = -1
    self._ping_interval = ping_interval
    self._last_ping_time = time()
    self._run_processes_each = run_processes_each
    self.threads_data = {}
    self.archived_threads = {}
    self._debug_state_dict = {
      'p_threads' : set(),
      'targets' : set()
    }
    self.process_tree = {}
    self.other_python_processes = []
    self.children_processes = []
    self._setup_thread()
    return
  
  
  def _setup_thread(self):
    self.P("Initializing {} v{} - monitoring prefix '{}' each {}s".format(
      self.__class__.__name__, self.__version__, self.monitored_prefix, self.tick
    ))
    if not hasattr(threading.enumerate()[0], 'native_id'):
      self.P("  SYSMON ERROR: Python {} does not support native thread IDs.".format(
        sys.version
      ), color='error')
    self.git_branch = self.log.git_branch
    self.os_name = self.log.get_os_name()
    self.python_version  = self.log.python_version
    self.conda_env = self.log.conda_env
    self.run_step()      
    return
    
  
  def P(self, s, **kwargs):
    self.log.P('[SYSMON] ' + s, **kwargs)
    return
  
  def P(self, s, color=None, **kwargs):
    if color is None or (isinstance(color,str) and color[0] not in ['e', 'r']):
      color = ct.COLORS.STATUS
    self.log.P('[SYSMON] ' + s, color=color,**kwargs)
    return      
  
  
  def __add_threads_info(self, threads):
    # first archive old threads
    all_threads = list(self.threads_data.keys())
    for thr in all_threads:
      if thr not in threads:
        # probably a old thread, lets archive it
        self.archived_threads[thr] = self.threads_data[thr].copy()
        del self.threads_data[thr]
    # now add or update threads
    for thr in threads:
      if thr not in self.threads_data:
        self.threads_data[thr] = {}
        # first get proc std attributes (non-ts)
        for fld in SysMonCt.P_ATTRS:
          if fld in threads[thr]:
            self.threads_data[thr][fld] = threads[thr][fld]
        # then create other non-timeseries attrs
        for fld in SysMonCt.NON_TIMESERIES:
          if fld in threads[thr]:
            self.threads_data[thr][fld] = threads[thr][fld]
      for k,v in threads[thr].items():        
        if k not in self.threads_data[thr]: 
          # if not create alread then must be a timeserie attr
          self.threads_data[thr][k] = deque(maxlen=self.max_size)
        if isinstance(self.threads_data[thr][k], deque):
          self.threads_data[thr][k].append(v)
        else:
          self.threads_data[thr][k] = v
    return
  

  ### PUBLIC

  @staticmethod
  def timestamp():
    return strftime('%Y-%m-%d %H:%M:%S', localtime(time()))

  @staticmethod
  def get_python_processes():
    targets = []
    piter = psutil.process_iter(SysMonCt.P_ATTRS + [SysMonCt.PID, SysMonCt.P_MEM])
    for proc in piter:
      if 'python' in proc.info[SysMonCt.NAME]:
        d = {k : v if v is not None else "" for k, v in proc.info.items() if k != SysMonCt.P_MEM}
        d[SysMonCt.P_MB] = proc.info[SysMonCt.P_MEM].rss // 1024**2
        d[SysMonCt.LAST_SEEN] = SystemMonitor.timestamp()
        targets.append(d)
    return targets
    
  
  def get_monitored_threads_info(self, prefix, proc=None, pid=None, interval=0.1, state_dict=None):
    dct_result = {}
    try:
      if proc is None:  
        p = psutil.Process(pid)    
      else:
        p = proc
      total_percent = p.cpu_percent(interval)
      p_times = p.cpu_times()
      total_time = sum(p_times)
      process_info = {
        p.pid : {
          SysMonCt.NAME : 'HOST_PROCESS',
          SysMonCt.P_PRC : total_percent,
          SysMonCt.UTIME : p_times.user,
          SysMonCt.P_TIME : total_time,
          SysMonCt.STIME : p_times.system,        
          SysMonCt.P_CUSR : p_times.children_user,        
          SysMonCt.P_CSYS : p_times.children_system,    
          SysMonCt.P_USR  : p.username(),
          SysMonCt.P_CWD  : p.cwd(),
          
          SysMonCt.LAST_SEEN : SystemMonitor.timestamp(),
        }
      }

      pst = p.threads()  
      thrs = threading.enumerate()
      threads = {}
      if hasattr(thrs[0], 'native_id'):
        prefix_len = len(prefix)
        targets = {x.native_id: x for x in thrs if prefix == x.name[:prefix_len]}
        if state_dict is not None and isinstance(state_dict, dict):
          state_dict['targets'].update(['{}:{}'.format(v.name,k) for k,v in targets.items()])
          state_dict['p_threads'].update([x.id for x in pst])
        threads = { 
          x.id : {
            SysMonCt.NAME : targets[x.id].name,
            SysMonCt.UTIME : x.user_time,
            SysMonCt.STIME : x.system_time,
            
            SysMonCt.LAST_SEEN : strftime('%Y-%m-%d %H:%M:%S', localtime(time()))
          }
          for x in pst if x.id in targets
        }    
      dct_result = {**process_info, **threads}
    except:      
      self.P("Error generating monitored threads info for pid: {}".format(pid), color='r')
      dct_result = None
    return dct_result  
  
  
  
  def display_process_tree_info(self, info=None):
    if info is None:
      info = self.get_process_tree()
    s = self.log.dict_pretty_format(info)
    self.log.P("Process tree:\n{}".format(s))
    return


  def get_process_tree(self, pid=None):
    """
    This is a one-shot process tree depth (recurse) analysis for a given 
    particular pid (current pid by default)

    Parameters
    ----------
    pid : int, optional
      the target process pid - current if None. The default is None.

    Returns
    -------
    dct_res : dict
      tree dict data structure containing process including child (recurse) 
      and threads.
      
    """
    def _generate_tree(dct_data, pid):
      try:
        p = psutil.Process(pid)
      except:
        self.P("Error generating process sub-tree for pid: {}".format(pid), color='r')
        return
      chlds = p.children()
      dct_info = p.as_dict(attrs=SysMonCt.P_ATTRS)
      for k,v in dct_info.items():
        dct_data[k] = " ".join(dct_info[k]) if isinstance(v, list) else v
      dct_data[SysMonCt.P_MB] = p.memory_info().rss // 1024**2
      for k,v in p.cpu_times()._asdict().items():
        dct_data[k] = v
      thrs = p.threads()
      if len(thrs) > 0:
        dct_data[SysMonCt.K_THR] = {}
        for thr in thrs:
          dct_data[SysMonCt.K_THR][thr.id] = {}
          for k,v in thr._asdict().items():
            if k != SysMonCt.K_ID:
              dct_data[SysMonCt.K_THR][thr.id][k] = v
              
      if len(chlds) > 0:
        dct_data[SysMonCt.K_CHL] = {}
        for i, chld in enumerate(chlds):
          dct_data[SysMonCt.K_CHL][chld.pid] = {}
          _generate_tree(dct_data[SysMonCt.K_CHL][chld.pid], chld.pid) 
      return
    #end _generate_tree
    dct_res = {}
    try:
      p = psutil.Process(pid)  
      dct_res = {p.pid : {}}
      _generate_tree(dct_res[p.pid], pid=p.pid)
    except:
      self.P("Error generating FULL process tree for pid: {}".format(pid), color='r')
    return dct_res


  def get_status_log(self, descending=True):
    """
    Generates the System Monitor status log as a list of strings

    Parameters
    ----------
    descending : bool, optional
      Sort the list of threads descending. The default is True.

    Returns
    -------
    buff : list
      list of strings.

    """
    buff = []    
    try:
      main_pid = self.main_pid 
      total_percent_last = self.threads_data[main_pid][SysMonCt.P_PRC][-1]
      total_percent_mean = np.mean(self.threads_data[main_pid][SysMonCt.P_PRC])
      total_time = self.threads_data[main_pid][SysMonCt.P_TIME][-1]
      user = self.threads_data[main_pid][SysMonCt.P_USR][:10]
      cwd = self.threads_data[main_pid][SysMonCt.P_CWD]
      git = self.git_branch
          
      buff.append('=' * 40 + ' System Monitor ' + '=' * 40)
      buff.append("Proc: {} [{}]  CPU mean/last: {:.1f}%/{:.1f}% ({:.1f}s)  Git:'{}'  Cwd: '{}'".format(
        main_pid, user, total_percent_mean, total_percent_last, total_time, 
        git, cwd)
      )

      lst_pids_and_times = [(
        _pid, 
        self.threads_data[_pid][SysMonCt.UTIME][-1] + self.threads_data[_pid][SysMonCt.STIME][-1]
        ) for _pid in self.threads_data]
      
      lst_pids_and_times = sorted(lst_pids_and_times, key=lambda x: x[1], reverse=descending)
      
      if len(lst_pids_and_times) == 1:
        buff.append("  Missing thread information:")
        buff.append("    - Maybe current process does not use ANY threads...")
        buff.append("    - Most likely wrong python version is installed. For linux please use py38 or above")
        buff.append("    - Detected OS: {}".format(self.os_name))
        buff.append("    - Detected Py: {}".format(self.python_version))
      
      for pid,thr_time in lst_pids_and_times:
        if pid == main_pid:
          continue
        name = self.threads_data[pid][SysMonCt.NAME]
        _utime = self.threads_data[pid][SysMonCt.UTIME][-1]
        _stime = self.threads_data[pid][SysMonCt.STIME][-1]
        thr_time1 = _utime + _stime
        time_is_ok = thr_time1 == thr_time
        if not time_is_ok:
          buff.append("  Something is WRONG with the time inside SysMon")
        local_perc = thr_time / (total_time + 1e-14)
        thr_overall_prc = total_percent_mean * local_perc
        _n = name[:36] + ' [{:<7}'.format(str(pid)+']')
        _n = _n + ' ' * (45 - len(_n[:45]))
        buff.append('  {} {:>4.1f}% (glb: {:>4.1f}%) Total: {:>4.1f} (stime: {:>5.2f} utime: {:>5.2f})'.format(
          _n, 
          local_perc * 100, thr_overall_prc, thr_time, 
          _stime, _utime)
        )
      if self._DEBUG:
        buff.append('')
        for k in self._debug_state_dict:
          ss = ", ".join([str(x) for x in sorted(self._debug_state_dict[k])])
          buff.append('  ## {:<10} {}'.format('{}:'.format(k), ss))

      if len(self.archived_threads) > 0:
        lst_archived = list(self.archived_threads.keys())
        n_archived = len(lst_archived)
        str_archived = ''
        for thr in lst_archived:
          str_archived += '{}...@{}, '.format(
            self.archived_threads[thr][SysMonCt.NAME][:10],
            self.archived_threads[thr][SysMonCt.LAST_SEEN]
          )
        buff.append('  Total {} archived threads\n'.format(n_archived))
        
      if len(self.children_processes) > 0:
        buff.append("Children processes:")
        for child in self.children_processes:
          chl_name = child[SysMonCt.NAME]
          buff.append("  PID:{:>20}  Mem:{:>5} MB  Name: '{}'  CWD: '{}'".format(
            "{} [{}]".format(child[SysMonCt.PID],child[SysMonCt.P_USR][:10]),
            child[SysMonCt.P_MB], chl_name,
            child[SysMonCt.P_CWD]
          ))
      
      if len(self.other_python_processes) > 0:
        buff.append("Other python processes:")
        for other_py in self.other_python_processes:
          try:
            other_pid = other_py[SysMonCt.PID]
            other_usr = other_py[SysMonCt.P_USR]
            if other_usr is not None and len(other_usr) > 10:
              other_usr = other_usr[:10]
            other_mb  = other_py[SysMonCt.P_MB]
            other_cwd = other_py[SysMonCt.P_CWD]
            buff.append("  PID:{:>20}  Mem:{:>5} MB  CWD: '{}'".format(
              "{} [{}]".format(other_pid, other_usr),
              other_mb, other_cwd          
            ))
          except:
            # do nothing if the data is not available - 
            self.P("Issue with python process analysis for: {}".format(other_py), color='r')
    except Exception as exc:
      buff.append("  EXCEPTION <{}> while generating sys-mon log.".format(str(exc)))
    return buff
  
  
  def log_status(self, full_log=False):
    """
    Outputs to log the current status of the System Monitor

    Parameters
    ----------
    full_log : bool, optional
      will display extended info if required. The default is False.

    Returns
    -------
    None.

    """
    buff = self.get_status_log()
    s = "SysMon info:\n" + buff[0]
    if len(buff) > 1:
      s += ':'
      for l in buff[1:]:
        s += '\n' + l 
    self.P(s)
    if full_log:
      s_tree = self.log.dict_pretty_format(self.process_tree)
      s_py = "\n".join([self.log.dict_pretty_format(x) for x in self.other_python_processes])
      self.P("Process tree:\n{}".format(s_tree))                       
      self.P("Other pythons running:\n{}".format(s_py))
    return
  
  
  def run_step(self):
    """
    Runs a full system analysis step

    Returns
    -------
    None.

    """
    self._loop_count += 1 # it starts at -1 :)
    self.log.start_timer('run_step', section='system_monitor')
    self.log.start_timer('get_monitored_threads_info', section='system_monitor')
    # extract all current threds within current process and update a simple debug state dict
    threads = self.get_monitored_threads_info(
      prefix=self.monitored_prefix,
      pid=self.main_pid,
      state_dict=self._debug_state_dict,
    )
    self.log.stop_timer('get_monitored_threads_info', section='system_monitor')
    if threads is not None:
      self.__add_threads_info(threads)
        
      if (self._loop_count % self._run_processes_each) == 0:
        self.log.start_timer('get_process_tree', section='system_monitor')
        # get current process tree
        self.process_tree = self.get_process_tree()
        self.log.stop_timer('get_process_tree', section='system_monitor')

        self.log.start_timer('get_python_processes', section='system_monitor')
        # get all python processes 
        python_processes = self.get_python_processes()
        # filter other running python(s)
        self.other_python_processes = [x for x in python_processes if x[SysMonCt.PID] != self.main_pid]
        # get current process children
        children = self.process_tree[self.main_pid].get(SysMonCt.K_CHL,{})
        self.children_processes = []
        if len(children) > 0:
          # copy children processes to a simple internal list
          for chl in children:
            children[chl][SysMonCt.PID] = chl
            self.children_processes.append(children[chl])
        self.log.stop_timer('get_python_processes', section='system_monitor')
      
      if (time() - self._last_ping_time) > self._ping_interval:
        self.P("System Monitor is alive.", color='g')
        self._last_ping_time = time()
    self.log.stop_timer('run_step', section='system_monitor')
    return  
  
  def start(self):
    self._thread = threading.Thread(
      target=self.run,
      name=self.name,
      daemon=True,
    )
    self._thread.start()
    return
   
    
  def run(self):
    """
    Main thread loop.

    Returns
    -------
    None.

    """
    self.P("Starting {} thread @ {}s monitoring interval...".format(
      self.__class__.__name__, self.tick), 
      color='g'
    )
    self.done = False
    while not self.done:
      self.run_step()
      sleep(self.tick)
    return
      
  
  def stop(self):
    self.P("Shutting down {}...".format(self.__class__.__name__), color='y')
    self.done = True
    self._thread.join()
    return
  
  
  def shutdown(self):
    self.stop()
    return


"""

Mar 01 08:38:13 k8s-m1 nerdctl[2450178]: Traceback (most recent call last):
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:   File "/opt/conda/lib/python3.10/site-packages/kmonitor/mixins/nodes_mixin.py", line 64, in get_nodes_metrics
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:     memory_usage_gib = int(node['usage']['memory'].rstrip('Ki')) / (1024**2)  # Convert KiB to GiB
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]: ValueError: invalid literal for int() with base 10: '6524M'
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]: During handling of the above exception, another exception occurred:
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]: Traceback (most recent call last):
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:   File "/exe_eng/core/business/base/base_plugin_biz_loop.py", line 234, in execute
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:     self._on_idle()
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:   File "/exe_eng/core/business/base/base_plugin_biz_loop.py", line 205, in _on_idle
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:     self.high_level_execution_chain()
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:   File "/exe_eng/core/business/base/base_plugin_biz_loop.py", line 314, in high_level_execution_chain
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:     self.process_wrapper()
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:   File "/exe_eng/core/business/base/base_plugin_biz.py", line 1607, in process_wrapper
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:     self._payload = self._process()
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:   File "/exe_eng/core/business/base/base_plugin_biz_api.py", line 63, in _process
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:     return self.process()
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:   File "/exe_eng/core/business/default/admin/k8s_monitor_01.py", line 147, in process
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:     dct_nodes = self.__get_status()
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:   File "/exe_eng/core/business/default/admin/k8s_monitor_01.py", line 111, in __get_status
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:     nodes = self.__km.get_nodes_metrics()
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:   File "/opt/conda/lib/python3.10/site-packages/kmonitor/mixins/nodes_mixin.py", line 79, in get_nodes_metrics
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:     self._handle_exception(exc)
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:   File "/opt/conda/lib/python3.10/site-packages/kmonitor/base.py", line 98, in _handle_exception
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]:     error_message += f"  Reason: {exc.reason}\n"
Mar 01 08:38:13 k8s-m1 nerdctl[2450178]: AttributeError: 'ValueError' object has no attribute 'reason'

"""
from naeural_core.business.base import BasePluginExecutor
try:
  import kmonitor
  K8S_PACKAGE_AVAIL = True
except:
  K8S_PACKAGE_AVAIL = False

FAKE_DATA =  [{
    "name": "k8s-m1-fake",
    "status": "Ready",
    "conditions": {
      "NetworkUnavailable": "False",
      "MemoryPressure": "False",
      "DiskPressure": "False",
      "PIDPressure": "False",
      "Ready": "True"
    },
    "cpu_usage_mili": 711,
    "memory_usage_gib": 7.64,
    "total_memory_gib": 15.61,
    "total_cpu_cores": 4,
    "memory_load": "48.9%",
    "cpu_load": "17.8%"
  },
  {
    "name": "k8s-w1-fake",
    "status": "Ready",
    "conditions": {
      "NetworkUnavailable": "False",
      "MemoryPressure": "False",
      "DiskPressure": "False",
      "PIDPressure": "False",
      "Ready": "True"
    },
    "cpu_usage_mili": 168,
    "memory_usage_gib": 3.05,
    "total_memory_gib": 7.75,
    "total_cpu_cores": 3,
    "memory_load": "39.3%",
    "cpu_load": "5.6%"
  },
  {
    "name": "k8s-w2-fake",
    "status": "Ready",
    "conditions": {
      "NetworkUnavailable": "False",
      "MemoryPressure": "False",
      "DiskPressure": "False",
      "PIDPressure": "False",
      "Ready": "True"
    },
    "cpu_usage_mili": 259,
    "memory_usage_gib": 3.25,
    "total_memory_gib": 7.75,
    "total_cpu_cores": 3,
    "memory_load": "41.9%",
    "cpu_load": "8.7%"
  }]



_CONFIG = {
  **BasePluginExecutor.CONFIG,

  'ALLOW_EMPTY_INPUTS'  : True,
  
  'PROCESS_DELAY'       : 5,
  

  'ALERT_DATA_COUNT'    : 1,
  'ALERT_RAISE_VALUE'   : 0.8,
  'ALERT_LOWER_VALUE'   : 0.79,
  'ALERT_MODE'          : 'min',
  
  'ARTIFICIAL'          : True,
  
  'DEBUG_MODE'          : False,

  'VALIDATION_RULES': {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES'],
  },
}



class K8sMonitor01Plugin(BasePluginExecutor):
  
  def __init__(self, **kwargs):
    super(K8sMonitor01Plugin, self).__init__(**kwargs)
    return
  
  
  def __get_fake_data(self):
    results = []
    for fake_node in FAKE_DATA:
      v = fake_node.copy()
      v['cpu_usage_mili'] = self.np.random.randint(0, v['total_cpu_cores'] * 1000)
      v['memory_usage_gib'] = self.np.random.uniform(0, v['total_memory_gib'])
      v['memory_load'] = f"{v['memory_usage_gib'] / v['total_memory_gib'] * 100:.1f}%"
      v['cpu_load'] = f"{v['cpu_usage_mili'] / v['total_cpu_cores'] / 1000 * 100:.1f}%"
      results.append(v)
    return results
      
  def __get_status(self):
    result = {
      'nodes': {},
      'pods': [],
      'namespaces': [],
    }
    if self.__km is None:
      if self.cfg_artificial:
        nodes = self.__get_fake_data()
        self.P("Sending artificial data for testing with {} nodes".format(len(nodes)))
    else:
      if hasattr(self.__km, 'summary'):
        result = self.__km.summary()
      else:
        nodes = self.__km.get_nodes_metrics()
        result['nodes'] = {x['name']: x for x in nodes}
    return result
  
  
   
  def on_init(self):
    is_ok = None
    self.__km = None
    self.__last_node_status = None
    if not K8S_PACKAGE_AVAIL:
      if self.cfg_artificial:
        s1 = "KubeMonitor package not available. Using artificial data for testing."
      else:
        s1 = "KubeMonitor package not available."  
    try:      
      self.__km = kmonitor.KubeMonitor(log=self.log)
      self.__last_node_status = self.__km.get_nodes_metrics()
      s1 = "KubeMonitor v{} initialized and monitoring {} nodes".format(
        kmonitor.__version__,
        len(self.__last_node_status)
      )
      is_ok = True
    except Exception as exc:
      self.__km = None
      if self.cfg_artificial:
        s1 = f"Failed to initialize KubeMonitor: {exc}. Using artificial data for testing."
      else:
        s1 = f"Failed to initialize KubeMonitor: {exc}"
    #end try
    if is_ok:
      self.P(s1, boxed=True)
    else:
      self.P(s1, color='r')
    self.__status = s1
    return
  
  
  def on_command(self, data, **kwargs):
    if isinstance(data, dict):
      cmd = data.get('cmd', None)
      args = data.get('args', None)
      if cmd == 'delete':
        if isinstance(args, list):
          pod_names = args[0]
          namespace = args[1] if len(args) > 1 else "default"
          self.P(f"Received delete pods command for wilcard '{pod_names}' on namespace {namespace}")
          self.__km.delete_pods_from_namespace(pod_names, namespace)
        #end if has args
      #end if cmd is delete
    return None
 
  
  def process(self):
    payload = None
    if self.is_supervisor_node:
      dct_result = self.__get_status()
      dct_nodes = dct_result.get('nodes', {})
      lst_pods = dct_result.get('pods', [])
      lst_namespaces = dct_result.get('namespaces', [])
      if len(dct_nodes) > 0:
        msg = "Node status for {} nodes: {}".format(len(dct_nodes), self.__status)
        payload = self._create_payload(
          k8s_nodes=dct_nodes,
          k8s_pods=lst_pods,
          k8s_namespaces=lst_namespaces,
          status=msg,
        )
    return payload


from naeural_core import constants as ct
from naeural_core.serving.base.base_serving_process import ModelServingProcess as BaseServingProcess

__VER__ = '0.1.0.0'

_CONFIG = {
  **BaseServingProcess.CONFIG,
  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },
}

class BasicTFSessionServer(BaseServingProcess):
  """
  This is a premade Tensorflow base serving process. You can subclass it in order
  to obtain advanced versions that preprocess or post-process
  """

  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(BasicTFSessionServer, self).__init__(**kwargs)
    return

  @property
  def cfg_gpu_min_memory_mb(self):
    return self.config_model.get(ct.GPU_MIN_MEMORY_MB)

  @property
  def _obsolete_cfg_memory_fraction(self):
    return self.config_model.get(ct.MEMORY_FRACTION)

  @property
  def cfg_graph(self):
    return self.config_model[ct.GRAPH]

  @property
  def cfg_url(self):
    return self.config_model.get(ct.URL, None)

  @property
  def cfg_input_tensors(self):
    return self.config_model.get(ct.INPUT_TENSORS, [])

  @property
  def cfg_output_tensors(self):
    return self.config_model.get(ct.OUTPUT_TENSORS, [])

  def _setup_session(self):
    self._start_timer('setup_session')
    import tensorflow.compat.v1 as tf
    mem_minimal = self.cfg_gpu_min_memory_mb
    cfg_mem_fraction = self._obsolete_cfg_memory_fraction
    total_gpu_mem_gb = self.log.gpu_info()[0]['TOTAL_MEM'] # assume in GB
    total_gpu_mem_mb = total_gpu_mem_gb * 1024

    if mem_minimal is None:
      self.P("{} not found in config. Please use it in place of memory percent.".format(
        ct.GPU_MIN_MEMORY_MB), color='r')
      if cfg_mem_fraction is None:
        raise ValueError("TF serving process must have either memory size im MB or (now obsolete) memory fraction")
      mem_fraction = cfg_mem_fraction
    else:
      self.P("  Model config {}={} MB".format(ct.GPU_MIN_MEMORY_MB, mem_minimal))    
      if cfg_mem_fraction is not None:
        self.P("Found both {} and {}. Using {}".format(
          ct.MEMORY_FRACTION, ct.GPU_MIN_MEMORY_MB, ct.GPU_MIN_MEMORY_MB))
      mem_fraction = mem_minimal / total_gpu_mem_mb
    #endif

    graph_file = self.cfg_graph
    self.P("Preparing '{}' from '{}' using {:.2f} memory fraction".format(
      self.server_name.upper(), graph_file, mem_fraction), color='y')

    url = self.cfg_url
    if url is not None:
      self.log.maybe_download_model(
        url=url,
        model_file=graph_file
      )
    #endif

    graph = self.log.load_graph_from_models(graph_file)
    if graph is None:
      raise ValueError("Graph loading failed for {}".format(graph_file))
    self.graph = graph
    self.P("Preparing session...")
    if mem_fraction is not None:
      gpu_options = tf.GPUOptions(
        per_process_gpu_memory_fraction=mem_fraction
        )
      config = tf.ConfigProto(gpu_options=gpu_options)
      self.sess = tf.Session(graph=self.graph, config=config)
      self.mem_fraction = mem_fraction
    else:
      self.sess = tf.Session(graph=self.graph)
      self.mem_fraction = 1
    self.tf_runoptions = tf.RunOptions(report_tensor_allocations_upon_oom=True)
    self._stop_timer('setup_session')
    return
  
  def _setup_tensors(self):
    self._start_timer('setup_tensors')
    input_tensor_names = self.cfg_input_tensors
    output_tensor_names = self.cfg_output_tensors
    self.P("Preparing input {} and output {} tensors".format(
      input_tensor_names, output_tensor_names))
    self.input_tensor_names = []
    self.output_tensor_names = []
    
    tensor_outputs = []
    for tname in output_tensor_names:
      tn = tname if ':' in tname else tname + ':0'
      tensor_outputs.append(self.sess.graph.get_tensor_by_name(tn))
      self.output_tensor_names.append(tn)
    self.tensor_outputs = tensor_outputs
    
    tensor_inputs = []
    for tname in input_tensor_names:
      tn = tname if ':' in tname else tname + ':0'
      tensor_inputs.append(self.sess.graph.get_tensor_by_name(tn))
      self.input_tensor_names.append(tn)
    self.tensor_inputs = tensor_inputs

    self._stop_timer('setup_tensors')
    msg = "\n  Status:\n    Inputs:\n{}    Outputs:\n{}    Mem prc: {:.1f}% ".format(
      "".join(["      {}\n".format(x) for x in self.tensor_inputs]), 
      "".join(["      {}\n".format(x) for x in self.tensor_outputs]),
      self.mem_fraction * 100 if hasattr(self, 'mem_fraction') else 100
      )
    return msg[:-1]
  
  def _pre_process_inputs(self, inputs):
    if len(self.input_tensor_names) > 1:
      self.P("Called {} with multi-inputs {} - please define a special server with specific `_preprocess_inputs`".format(
        self.__class__.__name__, self.input_tensor_names))
      raise ValueError("Using Basic TF server with multi-input graph not allowed")    
    dct_inputs = {
        self.input_tensor_names[0] : inputs['DATA']
      }    
    return dct_inputs
  
  def _run_inference(self, input_feed):
    fetches = self.tensor_outputs 
    if len(self.tensor_outputs) == 1:
      fetches = self.tensor_outputs[0]
    
    preds = self.sess.run(
      fetches=fetches,
      feed_dict=input_feed,
      options=self.tf_runoptions      
      )        
    return preds
  
  def _post_process_inference(self, predictions):
    return predictions
  
  def _filter_inference(self, predictions):
    return predictions
  
  def _post_process_outputs(self, predictions):
    return predictions

  def _prepare(self):
    return

  ###
  ### BELOW MANDATORY FUNCTIONS:
  ###
  
  def _startup(self):
    self._setup_session()
    msg = self._setup_tensors()
    self._prepare()
    return msg
  
  def _pre_process(self, inputs):
    prep_inputs = self._pre_process_inputs(inputs)

    if self.tensor_inputs:
      input_feed = {
        self.tensor_inputs[i] : prep_inputs[tensor_name]
        for i, tensor_name in enumerate(self.input_tensor_names)
      }
    else:
      input_feed = prep_inputs

    return input_feed

  
  def _predict(self, prep_inputs):
    preds = self._run_inference(prep_inputs)
    return preds  
  
  
  def _post_process(self, preds):
    self._start_timer(ct.TIMER_POST_PROCESS_INFERENCE)
    post_inference = self._post_process_inference(preds)
    self._stop_timer(ct.TIMER_POST_PROCESS_INFERENCE)
    
    self._start_timer(ct.TIMER_FILTER_RESULTS)
    filtered_inference = self._filter_inference(post_inference)
    self._stop_timer(ct.TIMER_FILTER_RESULTS)
    
    self._start_timer(ct.TIMER_POST_PROCESS_OUTPUTS)
    outputs = self._post_process_outputs(filtered_inference)
    self._stop_timer(ct.TIMER_POST_PROCESS_OUTPUTS)  
    return outputs
    
    
  
  





from naeural_core import Logger

if __name__ == '__main__':
  d = {'STREAM_NAME': 'TS_REST_AND_OTHER',
   'STREAM_METADATA': {'actual_dps': -1,
    'cap_resolution': 1,
    'cap_max_queue_len': 1,
    'cap_queue_len': 0,
    'cap_time': '2022-10-14 07:26:30.483792',
    'cap_elapsed_time': 0.06400084495544434,
    'live_feed': True,
    'reconnectable': 'YES'},
   'INPUTS': [{'IMG': None,
     'STRUCT_DATA': {'DUMMY DATA': 0},
     'METADATA': {'current_interval': None, 'meta_dummy_count': 1},
     'INIT_DATA': None,
     'TYPE': 'STRUCT_DATA'}],
   'SERVING_PARAMS': {}}
  
  l = Logger('DCT', base_folder='.', app_folder='_local_cache')
  l.dict_show(d)



from time import sleep
from threading import Thread
from naeural_core import Logger

from naeural_core.utils.sys_mon import SystemMonitor
                  

if __name__ == '__main__':
  l = Logger('THR', base_folder='.', app_folder='_cache')
  mon = SystemMonitor(
    log=l, 
    monitored_name='DecentrAI', 
    name='DecentrAI_MON',
    tick=0.1,
  )
  mon.start()

  done_flags = {
    1 : False,
    2 : False,
    }

  def stop_thread(thr):
    done_flags[thr] = True
    return
  
  data = ['start']
  data2 = ['start_bin']
  l.save_data_json(data, 'test.txt')
  l.save_pickle_to_data(
    data=data2,
    fn='test.pkl'
    )
  l.P(l._lock_table)  
  
  l.update_pickle_from_data(fn='test.pkl', update_callback=lambda x:x + ['update'])
  
    
  
  def t1():
    cnt = 0
    def t1_appender(lst):
      lst.append('t1_' + str(cnt))
      return lst
    for i in range(1000):
      # l.P("T1 " + str(i))
      cnt += 1
      l.update_data_json(fname='test.txt', update_callback=t1_appender)
      l.update_pickle_from_data(fn='test.pkl', update_callback=t1_appender)
      if done_flags[1]:
        l.P("Forcing quit T1")
        break
      else:
        sleep(0.001)
    return  


  def t2():
    cnt = 0
    def t2_appender(lst):
      lst.append('t2_' + str(cnt))
      return lst
    i = 0
    while True:
      # l.P("T2 " + str(i))
      cnt += 1
      i += 1
      l.update_data_json(fname='test.txt', update_callback=t2_appender)
      l.update_pickle_from_data(fn='test.pkl', update_callback=t2_appender)
      if done_flags[2]:
        l.P("Forcing quit T2")
        break
      else:
        sleep(0.001)
    return  
  
  thr1 = Thread(target=t1, name='DecentrAI_T1000', daemon=True)
  thr2 = Thread(target=t2, name='DecentrAI_TLoop', daemon=True)
  thr1.start()
  thr2.start()
  
  
  
  
  thr1.join(timeout=2)
  l.P("Thread thr1 join exited and thread is_alive = {}.".format(thr1.is_alive()))
  thr2.join(timeout=2)
  l.P("Thread thr2 join exited and thread is_alive = {}.".format(thr2.is_alive()))
  
  l.P("Stopping thr1")
  stop_thread(1)
  l.P("Stopping thr2")
  stop_thread(2)
  l.P("Thread thr2 is_alive={}".format(thr2.is_alive()))
  
  l.P(l._lock_table)
  
  d1 = l.load_data_json(fname='test.txt')
  d2 = l.load_pickle_from_data(fn='test.pkl')
  # l.P(d1)
  # l.P(d2)
  
  mon.stop()
  
  for line in mon.get_status_log():
    print(line)


from naeural_core.serving.default_inference.th_yf_base import YfBase as ParentServingProcess

_CONFIG = {
  **ParentServingProcess.CONFIG,

  "MODEL_WEIGHTS_FILENAME": "20240409_y8x_1152x2048_nms.ths",
  "MODEL_WEIGHTS_FILENAME_DEBUG": "20240409_y8x_1152x2048_nms_top6.ths",

  "URL": "minio:Y8/20240409_y8x_1152x2048_nms.ths",
  "URL_DEBUG": "minio:Y8/20240409_y8x_1152x2048_nms_top6.ths",

  'IMAGE_HW': (1152, 2048),

  'VALIDATION_RULES': {
    **ParentServingProcess.CONFIG['VALIDATION_RULES'],
  },
}

DEBUG_COVERED_SERVERS = False

__VER__ = '0.1.0.0'


class ThYf8x(ParentServingProcess):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(ThYf8x, self).__init__(**kwargs)
    return



#global dependencies
import numpy as np
import pandas as pd

#local dependenciess
from naeural_core import Logger
from naeural_core.local_libraries.vision.benchmark.model_benchmarker import ModelBenchmarker
# from naeural_core.local_libraries.nn.tf.utils import tfodapi2_ckpt_to_graph

if __name__ == '__main__':
  cfg_file = 'main_config.txt'
  log = Logger(
    lib_name='EE_BMRK', 
    config_file=cfg_file, 
    max_lines=1000, 
    TF_KERAS=False
    )
  
  # np_imgs = np.random.randint(
  #   low=0,
  #   high=256,
  #   size=(18, 720, 1280, 3),
  #   dtype=np.uint8
  #   )
  
  # #ED2_None_BS
  # l = []
  # for nr in [1, 2, 4, 8]:
  #   mb = ModelBenchmarker(
  #     log=log,
  #     module_name='xperimental.benchmarks.effdet.effdet2_None_models',
  #     class_name='ED2_None_BS{}'.format(nr)
  #     )
  #   df = mb.run(
  #     inputs=np_imgs,
  #     batch_sizes=[nr]
  #     )
  #   l.append(df)
  # df = pd.concat(l)
  # df.reset_index(drop=True, inplace=True)
  # log.p('\n\n{}'.format(df))
  
  
  # #ED2_1132x640
  # l = []
  # for nr in [1, 2, 4, 8]:
  #   mb = ModelBenchmarker(
  #     log=log,
  #     module_name='xperimental.benchmarks.effdet.effdet2_1132x640_models',
  #     class_name='ED2_1132x640_BS{}'.format(nr)
  #     )
  #   df = mb.run(
  #     inputs=np_imgs,
  #     batch_sizes=[nr]
  #     )
  #   l.append(df)
  # df = pd.concat(l)
  # df.reset_index(drop=True, inplace=True)
  # log.p('\n\n{}'.format(df))
  
  
  # l = []
  # for nr in [1, 2, 4, 8]:
  #   mb = ModelBenchmarker(
  #     log=log,
  #     module_name='xperimental.benchmarks.effdet.effdet4_None_models',
  #     class_name='ED4_None_BS{}'.format(nr)
  #     )
  #   df = mb.run(
  #     inputs=np_imgs,
  #     batch_sizes=[nr]
  #     )
  #   l.append(df)
  # df = pd.concat(l)
  # df.reset_index(drop=True, inplace=True)
  # log.p('\n\n{}'.format(df))
  
  
  
  
  

  
  
  
  
  
  


from naeural_core import constants as ct


class SharedMemoryManager(object):
  """
  Key-value store shared memory manager is a simple facade to a thread-safe data strcuture such 
  as a dict for any kind of objects that acts/works as a plugin:
    - allows the creation of relationship between various external actors (instaces of SHM)
    - basic functionalities such as getters & setters
    
  
  The dict-like of dict-based memory is organized as follows:
    
  'DEFAULT' : {
      'PLUGIN_1' : {
          ('STREAM_1', 'INSTANCE_1') : {
              'LINKED_INSTANCES' : [('STREAM_1', 'INSTANCE_2')]
              'SOME_KEY_OF_P1_I1' : 'SOME_VALUE'
          }
          ('STREAM_1', 'INSTANCE_2') : {
              'LINKED_INSTANCES' : None
              'LINKED_SERVER' : ('STREAM_1', 'INSTANCE_1')
              'SOME_KEY_OF_P1_I2' : 20
          }
      }
      'PLUGIN_2' : {
          ('STREAM_1', 'INSTANCE_1') : {
              'LINKED_INSTANCES' : [('STREAM_1', 'INSTANCE_2')]
              'SOME_KEY_OF_P2_I1' : 100
          }
          ('STREAM_1', 'INSTANCE_2') : {
              'LINKED_INSTANCES' : None
              'LINKED_SERVER' : ('STREAM_1', 'INSTANCE_1')
              'SOME_KEY_OF_P2_I2' : 'OTHER_VALUE'
          }
      }
  }  
      
  In above example we have two plugins that work on same stream and each has two instances
  with a simple server-client relationship
  
  """
  def __init__(
    self, 
    dct_shared : dict, 
    stream : str,
    plugin : str, 
    instance : str, 
    log : any, 
    category : str='DEFAULT', 
    linked_instances : 
      list =[]
    ):
    """
    Initialize the shared memory manager for a particular plugin type (signature) via self.reset(linked)

    Parameters
    ----------
    dct_shared : dict
        the shared memory dict
    stream : str
        name of the pipeline
    plugin : str
        signature of the plugin
    instance : str
        instance id of the plugin
    log : any
        Logger object
    category : str, optional
        Can be used for multiple categories, by default 'DEFAULT'
    linked_instances : list, optional
        default linked instances, by default []
    """
    assert isinstance(dct_shared, dict)
    super(SharedMemoryManager, self).__init__()
    self._categ = category # TODO: finish adding root category !! 
    self._shm = dct_shared
    self._plugin = plugin
    self._instance_key = (stream, instance)
    self.log = log    
    linked = self._process_linked_keys(linked_instances) # just a helper function
    self.reset(linked)
    return

  # Basic functions that can be overwritten with other shm approaches
  
  def P(self, s, color=None):
    s = f"[SHM{self._instance_key}] " + s
    self.log.P(s, color=color)
    return
  
  
  def get_global_dict(self):
    """
    Returns the global dict of the shared memory for the current category including all types of plugins

    Returns
    -------
    _type_
        _description_
    """
    if self._categ not in self._shm:
      self._shm[self._categ] = {}
    return self._shm[self._categ]
  
  
  def _maybe_init_shm(self, instance_key=None):
    """
    Maybe initialize the shared mem associated with current instance or other instance

    Parameters
    ----------
    instance_key : TYPE, optional
      Other instance than self if not None. The default is None.

    Returns
    -------
    None.

    """
    # We do not work directly with the global dict but with the plugin dict
    # so first we need to get the category (global dict)
    shm = self.get_global_dict()
    # then we need to get the plugin dict
    if self._plugin not in shm:
      shm[self._plugin] = {}
    if instance_key is not None:
      if instance_key not in shm[self._plugin]:
        shm[self._plugin][instance_key] = {}
    return
  
  def get_plugin_dict(self, instance_key=None):
    """
    Returns the dict of the current plugin instance or of a specified instance

    Parameters
    ----------
    instance_key : TYPE, optional
      Other instance than self if not None. The default is None.

    Returns
    -------
    dict
      The key:value store of the instance.

    """    
    shm = self.get_global_dict()
    self._maybe_init_shm(instance_key)
    if instance_key is not None:
      assert len(instance_key) == 2
      instance_key = tuple(instance_key)
      return shm[self._plugin][instance_key]
    return shm[self._plugin] 
  
  
  def get_instance_dict(self, instance_key : str =None):
    """
    Alias for get_plugin_dict

    Parameters
    ----------
    instance_key : str, optional
        the instance key, by default None
    """
    return self.get_plugin_dict(instance_key=instance_key)
  
  
  def set_local_key(self, key : str, val, instance_key=None):
    """
    Setter for a particular key-value

    Parameters
    ----------
    key : str
      the key.
    val : any
      the value that will be asigned to the key.
    instance_key : str, optional
      Other instance than self if not None. The default is None.

    Returns
    -------
    None.

    """
    if instance_key is None:
      instance_key = self.get_local_instance_key()
    self.get_plugin_dict(instance_key)[key] = val
    return


  def get_local_value(self, key : str, instance_key=None, default=None):
    """
    Getter for a local key-value in a instance key-value store

    Parameters
    ----------
    key : str
      the key.
    instance_key : str, optional
      Other instance than self if not None. The default is None[str].
    default : any, optional
      default value is key does not exists. The default is None.

    Returns
    -------
    TYPE
      DESCRIPTION.

    """
    if instance_key is None:
      instance_key = self.get_local_instance_key()    
    return self.get_plugin_dict(instance_key).get(key, default)  
  

  def get_local_instance_key(self):
    """
    Gets the id of the current manager object

    Returns
    -------
    TYPE
      DESCRIPTION.

    """
    return self._instance_key

  
  def get_all_instances_keys(self):
    """
    Gets all the keys of the current manager instance key-value store 

    Returns
    -------
    list of keys

    """
    return list(self.get_plugin_dict().keys())  

  
  def _process_linked_keys(self, linked_instances):
    """
    Basically helper function that returns the list of linked instances is the input is not a 
    normal list of linked instances but a string such as "MAIN" or "server".
    
    This function is OBSOLETE and will be removed soon. Please use `get_linked_keys()` approach instead
    
    Parses a linked_instances parameter that can be "MAIN", "server" or a list of instances ids. 
    Basically it receive a list of ("pipeline", "instance") and will set the current instance as
    a server and all other instances as clients
    

    Parameters
    ----------
    linked_instances : list or str
      list of ("pipeline", "instance") as the `signature` is already known.

    Returns
    -------
    res : TYPE
      DESCRIPTION.

    """
    res = linked_instances
    set_main = False
    if linked_instances is not None:
      if isinstance(linked_instances, list) and len(linked_instances) > 0:
        # first check if we have normal links        
        if isinstance(linked_instances[0], str) and linked_instances[0].upper() in [ct.MAIN, ct.server]:
          set_main = True
          self.P('WARNING: using `"LINKED_INSTANCES=["server"]` is deprecated and will be removed soon. Please use multiple pairs of ["STREAM","INSTANCE"]')
        else:
          res = [tuple(x) for x in linked_instances]
      elif (
        isinstance(linked_instances, str)
        and linked_instances.upper() in [ct.MAIN, ct.server]
      ):
        set_main = True
    if set_main:
      # Obsolete and will be removed soon
      self.set_local_key(key=ct.MAIN, val=True)
      res = self.get_main_links()
    return res
  
  
  # Higher level functions that do not use basic data types

  
  def is_main(self, instance_key : tuple = None):
    """
    This function checks if the current instance is a "main" instance or not. 
    This feature is OBSOLETE and will be removed soon. Please use `get_linked_keys()` instead
    """
    instance_key = self._instance_key if instance_key is None else instance_key
    return self.get_local_value(key=ct.MAIN, instance_key=instance_key)
  
  
  def get_local_dict(self):
    return self.get_plugin_dict(instance_key=self._instance_key)
  
  
  def get_linked_keys(self, instance_key : tuple = None, default : list = []):
    """
    Returns the linked instances of a particular instance or of the current instance
    Will return:
     - None if the instance is a client (default will not be returned as None was a valid configured value)
     - [] if the instance is single (default)
     - list of instances if the instance is a server
     
     TODO: Check/refactor this function

    Parameters
    ----------
    instance_key : tuple, optional
        the tuple key, by default None
    default : list, optional
        list of linked instances, by default []

    Returns
    -------
    _type_
        _description_
    """
    if instance_key is None:
      if self.is_main():
        # recheck & return
        # Obsolete and will be removed soon
        return self.get_main_links()
      instance_key = self.get_local_instance_key()
    res = self.get_local_value(
      key=ct.LINKED_INSTANCES, 
      instance_key=instance_key, 
      default=default
      )
    return res
  
  def get_linked_instances(self, instance_key : tuple = None, default : list = []):
    """
    Alias for get_linked_keys
    """
    return self.get_linked_keys(instance_key=instance_key, default=default)


  def _init_instance(self, instance_key : tuple , linked_instances : list = None):
    """
    Initialize a particular instance. 
    This function is used in the `reset` function and should not be used directly

    Parameters
    ----------
    instance_key : tuple
        the tuple key
    linked_instances : list, optional
        the list of linked instances, by default None

    Returns
    -------
    _type_
        _description_

    Raises
    ------
    ValueError
        _description_
    """
    self._maybe_init_shm(instance_key)
    # get already existing linked instances
    existing_linked_instances = self.get_linked_keys(instance_key=instance_key, default=[])
    if linked_instances is not None and len(linked_instances) > 0:
      if existing_linked_instances is None:
        # if existing_linked_instances was none then this instance used to be a client!
        if len(linked_instances) > 0:
          msg = "client-becoming-a-server modification detected for {}: received link proposal {} althogh it has been previously set to `None`".format(
            instance_key, linked_instances)
          self.P(msg, color='r')
    else:
      if existing_linked_instances is not None and len(existing_linked_instances) > 0:
        msg = "server-becoming-a-client modification detected for {}: received link proposal {} althogh it has been previously set to `{}`".format(
          instance_key, linked_instances, existing_linked_instances)
        self.P(msg, color='r')
    #endif check various scenarios
    self.set_linked(instance_key=instance_key, linked_instances=linked_instances)
    return self.get_plugin_dict(instance_key=instance_key)
  

  @property
  def linked(self):
    return self.get_linked_keys()
  

  def reset(self, linked_instances):
    """
    The main instance reset function. Will reset the current instance and all linked instances.    
    

    Parameters
    ----------
    linked_instances : list
        list of linked instances

    Raises
    ------
    ValueError
        _description_
    """
    # init whole plugin & instance
    self._init_instance(
      instance_key=self._instance_key,
      linked_instances=linked_instances
      )    
    # reset linked if any  
    local_instance = self.get_local_instance_key()
    if linked_instances is not None and len(linked_instances) > 0:      
      for instance in linked_instances:
        self._init_instance(instance_key=instance)
        
        # check for the client being a server
        other_instance_links = self.get_linked_keys(instance)
        if other_instance_links is not None and len(other_instance_links) > 0:
          msg = "Cross linked instances error (server over server): {} proposes link with {} while {} is a server as already references {}!".format(
            local_instance, linked_instances,
            instance, other_instance_links,
          )
          raise ValueError(msg)
        #end referecing a server
        
        # now check maybe this client already has a server
        existing_server = self.get_linked_server(instance_key=instance)
        if existing_server != local_instance and existing_server is not None and len(existing_server) > 0: # check for None as well as for a valid key
          msg = "Cross linked instances error (server overwrite attempt): {} proposes link with {} while {} already references  {}!".format(
            local_instance, linked_instances,
            existing_server, instance,
          )
          raise ValueError(msg)
        #end referecing a already taken client
        self.set_linked(instance, None)
        self.set_local_key(
          key=ct.LINKED_SERVER,
          val=local_instance,
          instance_key=instance,
          )
    else:
      # now we check to see if there are instances pointing to us as it seems we have no links
      all_instances = self.get_all_instances_keys()
      for instance in all_instances:
        server = self.get_linked_server(instance)
        if server == local_instance:
          self.P("Resetting server for {} as it was pointing to us".format(instance))
          self.set_local_key(
            key=ct.LINKED_SERVER,
            val=None, # set value to None so it resembles non existing key
            instance_key=instance,
          )
        #end referecing a client
    return
      
  
  def get_other_instances_keys(self):
    return list(set(self.get_all_instances_keys()) - set([self.get_local_instance_key()])) 
  
  
  def get_main_links(self): 
    """
    This function is OBSOLETE and will be removed soon. Please use `get_linked_keys()` instead
    """
    if self.is_main():
      # this is a main. will reset all other instances links, check for errors
      # and return all links
      instances = self.get_other_instances_keys()
      for instance in instances:
        if self.is_main(instance_key=instance):
          raise ValueError('Detected two "server" instances: {} and {}'.format(
            self.get_local_instance_key(), instance))
        else:
          self.set_linked(instance, None)
          self.set_local_key(
            key=ct.LINKED_SERVER,
            val=self.get_local_instance_key(),
            instance_key=instance,
            )
      return instances
    return None
  
  
  def get_linked_server(self, instance_key : tuple = None):
    """
    Returns the server of a particular instance or of the current instance

    Parameters
    ----------
    instance_key : tuple, optional
        the instance to be cheked, by default None

    Returns
    -------
    instance key of the server or None if the instance is a server
    """
    return self.get_local_value(
      instance_key=instance_key,
      key=ct.LINKED_SERVER, 
    )


  def set_linked(self, instance_key, linked_instances):
    """
    Sets the linked instances of a particular instance and also checks for errors
    """
    
    self.set_local_key(
      key=ct.LINKED_INSTANCES,
      val=linked_instances,
      instance_key=instance_key,
      )
    return

    
  def get_linked_data(self):
    """
    This will return the data of all linked instances. Please note that even if a certain linked instance
    does not exists this will return a basic initialized dict.

    Returns
    -------
    dict
        instance_key : dict of key:value
    """
    if self.linked is not None:
      if len(self.linked) > 0:
        return {k : self.get_plugin_dict(instance_key=k) for k in self.linked}
      # else:
      #   return {k : v for k, v in self.get_plugin_dict().items() 
      #           if k != self.get_local_instance_key()}
    return None  
  
  def clean_instance(self, instance_key : tuple = None):
    """
    WARNING: distructive function. Will remove the instance from the shared memory
    Cleans an instance from the shared memory
    """
    instance_key = self.get_local_instance_key() if instance_key is None else instance_key
    self.P("Cleaning instance {} from shared memory".format(instance_key))
    if instance_key in self.get_all_instances_keys():
      # first check if we are a server
      linked_instances = self.get_linked_keys(instance_key=instance_key)
      if linked_instances is not None and len(linked_instances) > 0:
        # we are a server so we need to clean the clients
        self.P("  Resetting linked instances: {}".format(linked_instances))
        for instance in linked_instances:
          self.set_local_key(
            key=ct.LINKED_SERVER,
            val=None,
            instance_key=instance,
          )
      self.get_plugin_dict().pop(instance_key)
    else:
      self.P("Instance {} not found in shared memory".format(instance_key))
    return
    
      
    
if __name__ == '__main__':
  
  from naeural_core import Logger
  l = Logger('SHMT', base_folder='.', app_folder='_cache')
  
  TESTS = [
    {
      'stream'            : 'STREAM_1',
      'plugin'            : 'PLUGIN_1',
      'instance'          : 'INSTANCE_1',
      'linked_instances'  : [
        ['STREAM_1','INSTANCE_2'],
        ['STREAM_1','INSTANCE_3'],
      ],
    },
    {
      'stream'            : 'STREAM_1',
      'plugin'            : 'PLUGIN_1',
      'instance'          : 'INSTANCE_2',
    },
    {
      'stream'            : 'STREAM_1',
      'plugin'            : 'PLUGIN_1',
      'instance'          : 'INSTANCE_3',
    },

    {
      'stream'            : 'STREAM_1',
      'plugin'            : 'PLUGIN_1',
      'instance'          : 'INSTANCE_4',
      'linked_instances'  : [
        ['STREAM_1','INSTANCE_5'],
        ['STREAM_1','INSTANCE_6'],
        # ['STREAM_1','INSTANCE_3'],
      ],
    },
    {
      'stream'            : 'STREAM_1',
      'plugin'            : 'PLUGIN_1',
      'instance'          : 'INSTANCE_5',
    },
    {
      'stream'            : 'STREAM_1',
      'plugin'            : 'PLUGIN_1',
      'instance'          : 'INSTANCE_6',
    },
    {
      'stream'            : 'STREAM_X',
      'plugin'            : 'PLUGIN_1',
      'instance'          : 'INSTANCE_A',
    },
    {
      'stream'            : 'STREAM_Y',
      'plugin'            : 'PLUGIN_1',
      'instance'          : 'INSTANCE_A',
    },
    {
      'stream'            : 'STREAM_Z',
      'plugin'            : 'PLUGIN_1',
      'instance'          : 'INSTANCE_A',
    },

  ]
  
  K_NAME = 'KEY_1'
  dct_shm = {}
  objs = []
  for i, test in enumerate(TESTS):
    obj = SharedMemoryManager(
      dct_shared=dct_shm, 
      log=l,
      **test,
    )  
    obj.set_local_key(key='KEY_1', val=10**i)
    objs.append(obj)
    
  def visit_status():
    for i, obj in enumerate(objs):
      linked = obj.get_linked_keys()
      is_server = linked is not None and len(linked) > 0
      is_standalone = linked == []
      server = obj.get_linked_server()
      is_client = server is not None
      l.P("Instance {}: ".format(obj.get_local_instance_key()))
      l.P("  Is server:   {}".format(is_server))
      l.P("  Is client:   {} {}".format(is_client, f"({server})" if is_client else ''))
      l.P("  Is stndalne: {}".format(is_standalone))
      l.P("  {}:       {}".format(K_NAME, obj.get_local_value(K_NAME)))
      if is_server:
        my_val = obj.get_local_value(K_NAME)
        linked_data = obj.get_linked_data()
        agg = None
        for instance in linked_data:
          if K_NAME not in linked_data[instance]:
            agg = "N/A due to missing {}".format(instance)
        if agg is None:
          others = sum(v[K_NAME] for k,v in linked_data.items())
          agg = my_val + others
        l.P("  Aggregated:  {}".format(agg))
    


  l.dict_pretty_format(d=dct_shm, display=True)
  
  l.P("Visiting instances", color='y')
  visit_status()

  if False:
    # play with first 3 intances
    l.P("Updating instance 1 & 2", color='y')
    # clean up
    objs[0] = SharedMemoryManager(
        dct_shared=dct_shm, 
        log=l,
        linked_instances=[],
        **{k:v for k,v in TESTS[0].items() if k != 'linked_instances'}, 
    )
    # modify the master
    objs[1] = SharedMemoryManager(
        dct_shared=dct_shm, 
        log=l,
        linked_instances=[['STREAM_1','INSTANCE_3']],
        **TESTS[1], # old idx 1 is now 0 and becomes master
    )  
  
  
  instance_id = -3
  objs[instance_id] = SharedMemoryManager(
      dct_shared=dct_shm, 
      log=l,
      linked_instances=[
        ['STREAM_Y','INSTANCE_A'],
      ],
      **{k:v for k,v in TESTS[instance_id].items() if k != 'linked_instances'}, 
  )  
  l.P("Revisiting instances", color='y')
  visit_status()

  instance_id = -3
  objs[instance_id] = SharedMemoryManager(
      dct_shared=dct_shm, 
      log=l,
      linked_instances=[
        ['STREAM_Y','INSTANCE_A'],
        ['STREAM_Z','INSTANCE_A'],
      ],
      **{k:v for k,v in TESTS[instance_id].items() if k != 'linked_instances'}, 
  )  
  l.P("Revisiting instances", color='y')
  visit_status()

      
  # l.P("Deleting instance 1", color='y')
  # # clean up
  # objs[0].clean_instance()
  # objs = objs[1:] # delete the first instance
  # # modify the master
  # objs[0] = SharedMemoryManager(
  #     dct_shared=dct_shm, 
  #     log=l,
  #     linked_instances=[['STREAM_1','INSTANCE_3']],
  #     **TESTS[1], # old idx 1 is now 0 and becomes master
  # )  
  # l.P("Revisiting instances"), color='y')
  # visit_status()



# 0xai_AhWU91s938BasILw7fyQ1bkGqWtw8dBCYDIg1Kz8wvfh MEUCIExlAnqAuMgJCPJ3GsguTE61MteZc0eR-akULK0vkIgRAiEA1PySc0xohCZGaX-C_kDilTQiUys0OFWniDT6WYpANJE=
# 0xai_AhWU91s938BasILw7fyQ1bkGqWtw8dBCYDIg1Kz8wvfh MEYCIQCUuHwnrzhrB2NlSJmY91n8vc1yBpnA0yrDmAgxE-Q8PgIhANgNKhcD3PJFLQaXSSfNdvC_PAbNz1YzWmHb-22O7aXE
data_message = ata_message = '{"2":100000,"3":3,"9":9,"10":{"1":1,"2":2,"100":100},"EE_SIGN":"MEUCIG3DqvsR_bqV6TqbJDwZ1ZvZSHgFvqhcHMCof4z08tSVAiEAiaD45YE4rsMYu4k1smFLI3rDBTk4W4QV2hclSb6TSag=","EE_SENDER":"0xai_AhWU91s938BasILw7fyQ1bkGqWtw8dBCYDIg1Kz8wvfh","EE_HASH":"714f7a182ab3e68bd819c8aaaa8da43d6c50b17d946e5ee1baceccd458afbd9d"}'
# Convert hex to bytes
import base64
import hashlib
import json
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.backends import default_backend

from time import time


dct_received = json.loads(data_message)
compressed_pk = dct_received['EE_SENDER'].replace('0xai_','')
signature_hex = dct_received['EE_SIGN']  
received_hash = dct_received['EE_HASH']
dct_data = {k:v for k,v in dct_received.items() if k not in ['EE_SIGN','EE_SENDER','EE_HASH']}
inputs = dct_data

if isinstance(inputs,str):
  data = inputs.encode()
elif isinstance(inputs,dict):  
  deterministic_json = json.dumps(inputs, sort_keys=True, separators=(',',':'))
  data = deterministic_json.encode()


# direct
obj_hash = hashlib.sha256(data)
str_hash = obj_hash.hexdigest()
b_hash = obj_hash.digest()

signatureB64 = signature_hex
signature_bytes = base64.urlsafe_b64decode(signatureB64)

# Compressed point
t0 = time()
t01 = time()
bpublic_key = base64.urlsafe_b64decode(compressed_pk)
e11 = time() - t01
t02 = time()
crv = ec.SECP256K1()
e12 = time() - t02 
t03 = time()
pk = ec.EllipticCurvePublicKey.from_encoded_point(
  curve=crv, 
  data=bpublic_key
)
e13 = time() - t03
e1 = time() - t0
# RECODING: To get the hexadecimal representation of the public key
public_bytes = pk.public_bytes(
    encoding=ec._serialization.Encoding.X962,
    format=ec._serialization.PublicFormat.CompressedPoint
)  
str_recoded_addr = base64.urlsafe_b64encode(public_bytes).decode()
assert str_recoded_addr == compressed_pk, "Recoding failed"
print("Recoding passed")

assert received_hash == str_hash, "Hashes do not match"
print("Hashing match passed")

try:
  t1 = time()
  pk.verify(signature_bytes, b_hash, ec.ECDSA(hashes.SHA256()))
  e2 = time() - t1
  elapsed_total = e1 + e2
  print("X962/Compres pk signature check valid in {:.3f}s ({:.3f}s ({:.3f}s+{:.3f}s+{:.3f}s) + {:.3f}s)".format(
      elapsed_total, e1, e11, e12, e13,
      e2
    )
  )
except Exception as e:
  print(f"ERROR: Invalid signature: {e}")
    



AI_ENGINES = {}

AI_ENGINES['general_detector'] = {
  # 'SERVING_PROCESS': 'th_y5l6s'
  'SERVING_PROCESS': 'th_yf8l'  # Y8L best so far for large res
}

AI_ENGINES['lowres_general_detector'] = {
  # 'SERVING_PROCESS': 'th_y5s6s'
  # 'SERVING_PROCESS': 'th_yf5s'  # y5s best so far for low res better than y8s as it has better input resolution
  'SERVING_PROCESS': 'th_yf8s'
}

AI_ENGINES['nano_general_detector'] = {
  'SERVING_PROCESS': 'th_yf8n'
}

AI_ENGINES['general_detector_xp'] = {
  'SERVING_PROCESS': 'th_yf_xp',
}

AI_ENGINES['yf5_detector'] = {
  'SERVING_PROCESS': 'th_yf5l'
}

AI_ENGINES['lowres_yf5_detector'] = {
  'SERVING_PROCESS': 'th_yf5s'
}

# AI_ENGINES['yf8_detector'] = {
#   'SERVING_PROCESS': 'th_yf8l'
# }

AI_ENGINES['lowres_yf8_detector'] = {
  'SERVING_PROCESS': 'th_yf8s'
}

AI_ENGINES['advanced_general_detector'] = {
  # 'SERVING_PROCESS': 'th_effdet7'
  'SERVING_PROCESS': 'th_yf8x'
}

AI_ENGINES['pose_detector'] = {
  # 'SERVING_PROCESS' : 'th_y5l6s_move',
  # 'SERVING_PROCESS': 'th_y5l6s_mspn',
  'SERVING_PROCESS': 'th_yf_mspn'
}

AI_ENGINES['lowres_pose_detector'] = {
  # 'SERVING_PROCESS': 'th_y5s6s_mspn'
  'SERVING_PROCESS': 'th_yfs_mspn'
}

AI_ENGINES['nano_pose_detector'] = {
  'SERVING_PROCESS': 'th_yfn_mspn'
}

AI_ENGINES['face_detector'] = {
  'SERVING_PROCESS': 'th_rface'
}

AI_ENGINES['face_detector_identification'] = {
  'SERVING_PROCESS': 'th_retina_face_resnet_identification'  # 'face_identification'
}

AI_ENGINES['face_id'] = {
  'SERVING_PROCESS': 'th_rface_id'  # 'face_identification'
}

AI_ENGINES['lowres_face_detector'] = {
  'SERVING_PROCESS': 'th_rface_s'
}

AI_ENGINES['custom_second_stage_detector'] = {
  # TODO: support seems to be lost downstream
  'SERVING_PROCESS': 'th_yolo_second_stage',
  'REQUIRES_INSTANCE': True  # requires ('th_yolo_second_stage', 'safety_helmet') style
}

# TODO:
#  below config is just for demo purposes
#  so this file should be split between "core" and "plugins"
AI_ENGINES['a_dummy_ai_engine'] = {
  'SERVING_PROCESS': 'a_dummy_classifier',
  'PARAMS': {
    'TEST_INFERENCE_PARAM': 1,  # should overwrite default 0
    'TEST_ENGINE_PARAM': 100,
  }
}

AI_ENGINES['a_dummy_cv_ai_engine'] = {
  'SERVING_PROCESS': 'a_dummy_cv_classifier',
  'PARAMS': {
    'TEST_INFERENCE_PARAM': 1,  # should overwrite default 0
    'TEST_ENGINE_PARAM': 100,
  }
}


AI_ENGINES['th_training'] = {
  'SERVING_PROCESS': 'th_training',
  'REQUIRES_INSTANCE': True
}

try:
  from extensions.serving.ai_engines.stable import AI_ENGINES as EXT_AI_ENGINES
  AI_ENGINES = {
    **AI_ENGINES,
    **EXT_AI_ENGINES
  }
except ImportError:
  import warnings
  msg = """
No extensions found for AI_ENGINES. Using only the default engines.
To add custom engines, create a file at `extensions/serving/ai_engines/stable.py` in your project and define 
a dictionary named `AI_ENGINES` following the structure in the default file. Alternatively, use 
the file name of your custom serving engine as the AI_ENGINE name, without needing to define `AI_ENGINES`.
To suppress this warning, set the environment variable `PYTHONWARNINGS` to 'ignore::UserWarning'.
  """
  warnings.warn(msg, UserWarning)





import torch as th


if __name__ == '__main__':
  shape = (4, 4, 4, 4, 2)
  inp = th.randn(shape)
  dropout_layer = th.nn.Dropout(p=0.5)
  flatten_layer = th.nn.Flatten()
  th.manual_seed(0)
  inp_drop = dropout_layer(inp)
  inp_drop_flatten = flatten_layer(inp_drop)

  inp_flatten = flatten_layer(inp)
  th.manual_seed(0)
  inp_flatten_drop = dropout_layer(inp_flatten)

  print(th.abs(inp_drop_flatten - inp_flatten_drop))



from .base_plugin_biz import BasePluginExecutor
from .cv_plugin_executor import CVPluginExecutor
from .sd_plugin_executor import BasicStructuredDataPluginExecutor
from .simple_rest_executor import SimpleRestExecutor
from .network_processor import NetworkProcessorPlugin



def minio_download(local_filename,
                   endpoint, 
                   access_key, 
                   secret_key, 
                   bucket_name, 
                   object_name):
  from minio import Minio
  from minio.commonconfig import GOVERNANCE

  try:
    client = Minio(
        endpoint=endpoint,
        access_key=access_key,
        secret_key=secret_key,
        secure=False,
        )
    
    res = client.fget_object(
      bucket_name=bucket_name, 
      object_name=object_name, 
      file_path=local_filename,
      )
  except Exception as e:
    print(e) # modifiy in self.P(e, color='error')
    return None  

  return local_filename

def minio_upload(file_path, 
                 endpoint, 
                 access_key, 
                 secret_key, 
                 bucket_name, 
                 object_name=None,
                 days_retention=None,
                 debug=False,
                 ):
  from minio import Minio
  from minio.commonconfig import GOVERNANCE
  from minio.retention import  Retention 
  from datetime import datetime as dttm
  from datetime import timedelta as tdelta
  from uuid import uuid4
  
  if object_name is None:
    object_name = "OBJ_"+ str(uuid4()).upper().replace('-','')
  
  try:
    client = Minio(
        endpoint=endpoint,
        access_key=access_key,
        secret_key=secret_key,
        secure=False,
        )
    
    retention = None
    if days_retention is None:
      date = dttm.utcnow().replace(
        hour=0, minute=0, second=0, microsecond=0,
        ) + tdelta(days=days_retention)
      retention = Retention(GOVERNANCE, date)
        
    result = client.fput_object(
      file_path=file_path,
      bucket_name=bucket_name,
      object_name=object_name,
      retention=retention,
      )
    
    url = client.presigned_get_object(
      bucket_name=result.bucket_name, 
      object_name=result.object_name,
      )
    
    print("Uploaded '{}' as '{}'".format(file_path, url))
  except Exception as e:
    print(e) # modifiy in self.P(e, color='error')
    return None
  
  return url, object_name
  
if __name__ == '__main__':
  
  fn = 'xperimental/_images/H480_W640/faces1.jpg'
  endpoint = 'MINIO_ENDPOINT'
  access_key = 'MINIO_ACCESS_KEY'
  secret_key = 'MINIO_SECRET_KEY'
  bucket_name ="MINIO_BUCKET_NAME"
  object_name = None
  days_retention = 5
  
  
  url, object_name = minio_upload(
    file_path=fn, 
    endpoint=endpoint, 
    access_key=access_key, 
    secret_key=secret_key, 
    bucket_name=bucket_name, 
    days_retention=days_retention,
    )
  
  down = minio_download(
    local_filename='xperimental/minio/test.jpg',
    endpoint=endpoint, 
    access_key=access_key, 
    secret_key=secret_key, 
    bucket_name=bucket_name,   
    object_name=object_name,
    )
  
  
  
 